# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""Parameters for neutron interactions with atoms."""

from __future__ import annotations

import dataclasses
import importlib.resources
from functools import lru_cache
from typing import TextIO

import scipp as sc


def reference_wavelength() -> sc.Variable:
    """Return the reference wavelength for absorption cross-sections.

    Returns
    -------
    :
        1.7982 Å
    """
    return sc.scalar(1.7982, unit='angstrom')


@dataclasses.dataclass(frozen=True, eq=False)
class ScatteringParams:
    """Scattering parameters for neutrons with a specific element / isotope.

    Provides access to the scattering lengths and cross-sections of neutrons
    with a given element or isotope.
    Values have been retrieved at 2024-02-19T17:00:00Z from the list at
    https://www.ncnr.nist.gov/resources/n-lengths/list.html
    which is based on :cite:`sears:1992`.
    Values are ``None`` where the table does not provide values.

    The absorption cross-section applies to neutrons with a wavelength
    of 1.7982 Å.
    See :func:`reference_wavelength`.
    """

    isotope: str
    """Element / isotope name."""
    coherent_scattering_length_re: sc.Variable | None = None
    """Bound coherent scattering length (real part)."""
    coherent_scattering_length_im: sc.Variable | None = None
    """Bound coherent scattering length (imaginary part)."""
    incoherent_scattering_length_re: sc.Variable | None = None
    """Bound incoherent scattering length (real part)."""
    incoherent_scattering_length_im: sc.Variable | None = None
    """Bound incoherent scattering length (imaginary part)."""
    coherent_scattering_cross_section: sc.Variable | None = None
    """Bound coherent scattering cross-section."""
    incoherent_scattering_cross_section: sc.Variable | None = None
    """Bound incoherent scattering cross-section."""
    total_scattering_cross_section: sc.Variable | None = None
    """Total bound scattering cross-section."""
    absorption_cross_section: sc.Variable | None = None
    """Absorption cross-section for λ = 1.7982 Å neutrons."""

    def __eq__(self, other: object) -> bool | type(NotImplemented):
        if not isinstance(other, ScatteringParams):
            return NotImplemented
        return all(
            self.isotope == other.isotope
            if field.name == 'isotope'
            else _eq_or_identical(getattr(self, field.name), getattr(other, field.name))
            for field in dataclasses.fields(self)
        )

    @staticmethod
    @lru_cache
    def for_isotope(isotope: str) -> ScatteringParams:
        """Return the scattering parameters for the given element / isotope.

        Parameters
        ----------
        isotope:
            Name of the element or isotope.
            For example, 'H', '3He', 'V', '50V'.

        Returns
        -------
        :
            Neutron scattering parameters.
        """
        with _open_scattering_parameters_file() as f:
            while line := f.readline():
                name, rest = line.split(',', 1)
                if name == isotope:
                    return _parse_line(isotope, rest)
        raise ValueError(f"No entry for element / isotope '{isotope}'")


def _open_scattering_parameters_file() -> TextIO:
    return (
        importlib.resources.files('scippneutron.atoms')
        .joinpath('scattering_parameters.csv')
        .open('r')
    )


def _parse_line(isotope: str, line: str) -> ScatteringParams:
    line = line.rstrip().split(',')
    return ScatteringParams(
        isotope=isotope,
        coherent_scattering_length_re=_assemble_scalar(line[0], line[1], 'fm'),
        coherent_scattering_length_im=_assemble_scalar(line[2], line[3], 'fm'),
        incoherent_scattering_length_re=_assemble_scalar(line[4], line[5], 'fm'),
        incoherent_scattering_length_im=_assemble_scalar(line[6], line[7], 'fm'),
        coherent_scattering_cross_section=_assemble_scalar(line[8], line[9], 'barn'),
        incoherent_scattering_cross_section=_assemble_scalar(
            line[10], line[11], 'barn'
        ),
        total_scattering_cross_section=_assemble_scalar(line[12], line[13], 'barn'),
        absorption_cross_section=_assemble_scalar(line[14], line[15], 'barn'),
    )


def _assemble_scalar(value: str, std: str, unit: str) -> sc.Variable | None:
    if not value:
        return None
    value = float(value)
    variance = float(std) ** 2 if std else None
    return sc.scalar(value, variance=variance, unit=unit)


def _eq_or_identical(a: sc.Variable | None, b: sc.Variable | None) -> bool:
    if a is None:
        return b is None
    return sc.identical(a, b)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import dataclasses
import uuid
from string import Template
from typing import TYPE_CHECKING

import scipp as sc

from ._resources import disk_chopper_repr_template, disk_chopper_style

try:
    from scipp.visualization.formatting_datagroup_html import (
        _datagroup_detail,
        _format_shape,
        load_dg_repr_tpl,
    )
except ImportError:
    from scipp.html.formatting_datagroup_html import (
        _datagroup_detail,
        _format_shape,
        load_dg_repr_tpl,
    )

if TYPE_CHECKING:
    from ..chopper import DiskChopper
else:
    DiskChopper = object


def disk_chopper_html_repr(chopper: DiskChopper) -> str:
    field_repr = _datagroup_repr(
        sc.DataGroup(
            {
                field.name: getattr(chopper, field.name)
                for field in dataclasses.fields(chopper)
                if not field.name.startswith('_')
            }
        ),
        type_name="DiskChopper",
        max_length_before_fold=16,
    )
    try:
        image = chopper.make_svg()
    except RuntimeError:
        image = ""
    return disk_chopper_repr_template().substitute(
        style_sheet=disk_chopper_style(), fields=field_repr, image=image
    )


def _datagroup_repr(
    dg: sc.DataGroup, *, type_name: str, max_length_before_fold: int = 15
) -> str:
    checkbox_status = "checked" if len(dg) < max_length_before_fold else ''
    header_id = "group-view-" + str(uuid.uuid4())
    details = _datagroup_detail(dg)
    html = Template(load_dg_repr_tpl())
    return html.substitute(
        style_sheet='',
        header_id=header_id,
        checkbox_status=checkbox_status,
        obj_type=type_name,
        shape_repr=_format_shape(dg, br_at=200),
        details=details,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

import json
from functools import partial, reduce
from pathlib import Path
from typing import TYPE_CHECKING

import plopp as pp
import scipp as sc
from matplotlib.colors import to_rgb
from plopp.core.typing import FigureLike

if TYPE_CHECKING:
    from mpltoolbox import Patch


def _define_shape_mask(da: sc.DataArray, info: dict) -> sc.Variable:
    """
    Function that creates a mask inside the area
    covered by the shape.
    """
    bounds = info["bounds"]
    return reduce(
        lambda a, b: a & b,
        [
            (da.coords[dim] >= lims["min"]) & (da.coords[dim] <= lims["max"])
            for dim, lims in bounds.items()
        ],
    )


def _get_rect_info(artist: Patch, figure: FigureLike) -> dict:
    """
    Convert the raw rectangle info to a dict containing the dimensions of
    each axis, and values with units.
    """
    x1 = artist.xy[0]
    x2 = artist.xy[0] + artist.width
    y1 = artist.xy[1]
    y2 = artist.xy[1] + artist.height
    return lambda: {
        "kind": "rectangle",
        "bounds": {
            figure.canvas.dims["x"]: {
                "min": sc.scalar(min(x1, x2), unit=figure.canvas.units["x"]),
                "max": sc.scalar(max(x1, x2), unit=figure.canvas.units["x"]),
            },
            figure.canvas.dims["y"]: {
                "min": sc.scalar(min(y1, y2), unit=figure.canvas.units["y"]),
                "max": sc.scalar(max(y1, y2), unit=figure.canvas.units["y"]),
            },
        },
    }


def _get_vspan_info(artist: Patch, figure: FigureLike) -> dict:
    x1 = artist.left
    x2 = artist.right
    return lambda: {
        "kind": "vspan",
        "bounds": {
            figure.canvas.dims["x"]: {
                "min": sc.scalar(min(x1, x2), unit=figure.canvas.units["x"]),
                "max": sc.scalar(max(x1, x2), unit=figure.canvas.units["x"]),
            }
        },
    }


def _get_hspan_info(artist: Patch, figure: FigureLike) -> dict:
    y1 = artist.bottom
    y2 = artist.top
    return lambda: {
        "kind": "hspan",
        "bounds": {
            figure.canvas.dims["y"]: {
                "min": sc.scalar(min(y1, y2), unit=figure.canvas.units["y"]),
                "max": sc.scalar(max(y1, y2), unit=figure.canvas.units["y"]),
            }
        },
    }


def _apply_masks(da: sc.DataArray, *masks: sc.Variable) -> sc.DataArray:
    out = da.copy(deep=False)
    for i, mask in enumerate(masks):
        out.masks[str(i)] = mask
    return out


def _scalar_to_dict(scalar: sc.Variable) -> dict:
    return {"value": scalar.value, "unit": str(scalar.unit)}


class MaskingTool:
    def __init__(self, data: sc.DataArray, color='magenta', **kwargs):
        """
        Interactive masking tool for 1D and 2D data.
        The tool will display a figure with the data and allow the user to
        draw rectangles, horizontal spans, and vertical spans to create masks, using
        buttons in the top bar of the figure.

        Instructions:

        - Use the buttons in the top bar to add masks to the data
        - Left-click to add a new shape, and left-click again to persist the shape
        - Left-click a vertex to edit a shape
        - Right-click and hold to drag a shape
        - Middle-click (or Ctrl + left-click) to delete a shape
        - Save the masks to a file when the "Save" button is clicked

        Parameters
        ----------
        data:
            The data to be masked.
        color:
            The color of the shapes drawn by the user.
        kwargs:
            Additional keyword arguments passed to the figure constructor.
        """
        import ipywidgets as ipw
        from mpltoolbox import Hspans, Rectangles, Vspans
        from plopp.plotting.common import require_interactive_figure
        from plopp.widgets import DrawingTool, style

        # Convert potential bin edge coords to midpoints
        da = data.copy(deep=False)
        for dim, coord in data.coords.items():
            if data.coords.is_edges(dim):
                da.coords[dim] = sc.midpoints(coord)

        ndim = da.ndim
        figure_lib = {1: pp.linefigure, 2: pp.imagefigure}
        if ndim not in figure_lib:
            raise ValueError("The masking tool can only handle 1D or 2D data.")
        self.data_node = pp.Node(da)
        self.masking_node = pp.Node(_apply_masks, self.data_node)
        self.fig = figure_lib[ndim](self.masking_node, **kwargs)
        require_interactive_figure(self.fig, "Masking tool")

        common = {
            "figure": self.fig,
            "input_node": self.data_node,
            "destination": self.masking_node,
        }

        col_args = {"edgecolor": color, "facecolor": (*to_rgb(color), 0.05)}

        rects = DrawingTool(
            tool=partial(Rectangles, **col_args),
            get_artist_info=_get_rect_info,
            icon="vector-square",
            func=_define_shape_mask,
            tooltip="Add rectangular masks",
            disabled=ndim == 1,
            **common,
        )
        vspans = DrawingTool(
            tool=partial(Vspans, **col_args),
            get_artist_info=_get_vspan_info,
            icon="grip-lines-vertical",
            func=_define_shape_mask,
            tooltip="Add vertical masks",
            **common,
        )
        hspans = DrawingTool(
            tool=partial(Hspans, **col_args),
            get_artist_info=_get_hspan_info,
            icon="grip-lines",
            func=_define_shape_mask,
            tooltip="Add horizontal masks",
            disabled=ndim == 1,
            **common,
        )
        self.controls = [rects, vspans, hspans]

        self.fig.top_bar.add(
            ipw.Label(
                "Add masks:",
                layout={
                    "display": "flex",
                    "justify_content": "flex-end",
                    "width": "110px",
                },
            )
        )
        for c in self.controls:
            self.fig.top_bar.add(c)
            c._tool._on_change.clear()
            c._tool.on_vertex_release(c.update_node)
            c._tool.on_drag_release(c.update_node)
            c._tool.on_remove(self.masking_node.notify_children)
            c.observe(self.toggle_button_states, names="value")

        self.filename = ipw.Text(placeholder="Save to file", layout={"width": "200px"})
        self.filename.observe(self.validate_filename, names="value")
        self.save_button = ipw.Button(
            icon="save", tooltip="Save to JSON", disabled=True, **style.BUTTON_LAYOUT
        )
        self.save_button.on_click(self._save_button_click)
        self.toggle_visibility = ipw.ToggleButton(
            value=True, icon="eye", tooltip="Hide shapes", **style.BUTTON_LAYOUT
        )
        self.toggle_visibility.observe(self.toggle_shape_visibility, names="value")
        self.fig.top_bar.add(ipw.HBox([], layout={"width": "40px"}))
        self.fig.top_bar.add(self.toggle_visibility)
        self.fig.top_bar.add(self.filename)
        self.fig.top_bar.add(self.save_button)

        self.info = ipw.Button(
            icon="question-circle",
            tooltip="""
Instructions:
• Use the buttons in the top bar to add masks to the data
• Left-click to add a new shape, and left-click again to persist the shape
• Left-click a vertex to edit a shape
• Right-click and hold to drag a shape
• Middle-click (or Ctrl + left-click) to delete a shape
• Save the masks to a file when the "Save" button is clicked""",
            **style.BUTTON_LAYOUT,
        )
        self.fig.top_bar.add(self.info)

    def toggle_button_states(self, change: dict) -> None:
        if change["new"]:
            for c in self.controls:
                if c.value and c is not change["owner"]:
                    c.value = False

    def validate_filename(self, change: dict) -> None:
        self.save_button.disabled = not change["new"]

    def toggle_shape_visibility(self, change: dict):
        for c in self.controls:
            for child in c._tool.children:
                child.set(visible=change["new"])
        self.toggle_visibility.icon = "eye" if change["new"] else "eye-slash"
        self.toggle_visibility.tooltip = (
            "Hide shapes" if change["new"] else "Show shapes"
        )
        self.fig.canvas.draw()

    def get_masks(self) -> dict:
        masks = {}
        mask_counter = 0
        for c in self.controls:
            for node in c._draw_nodes.values():
                info = node()
                mask_dims = "".join(info["bounds"].keys())
                mask_name = f"{mask_dims}_{mask_counter}"
                masks[mask_name] = {
                    "kind": info["kind"],
                    "bounds": {
                        dim: {
                            "min": _scalar_to_dict(lims["min"]),
                            "max": _scalar_to_dict(lims["max"]),
                        }
                        for dim, lims in info["bounds"].items()
                    },
                }
                mask_counter += 1
        return masks

    def _save_button_click(self, _=None) -> None:
        self.save_masks(filename=self.filename.value)

    def save_masks(self, filename: Path | str) -> None:
        """
        Save the masks to a JSON file.

        Parameters
        ----------
        filename:
            The name of the file to save the masks to. The file extension
            will be automatically added if not present.
        """
        with open(str(filename).removesuffix(".json") + ".json", "w") as f:
            json.dump(self.get_masks(), f, indent=2)

    def _repr_mimebundle_(self, **kwargs) -> dict:
        return self.fig._repr_mimebundle_()



import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_allclose
from scipy.integrate import quad

from scippneutron.absorption import Cylinder, Material, compute_transmission_map
from scippneutron.atoms import ScatteringParams


def transmission_fraction_case1(effective_attenuation_factor):
    '''Transmission fraction at a point on the z-axis
    infinitely far away when the sample is a cylinder
    symmetrical around y, with 1mm radius and 1mm height.
    '''
    return (
        2
        / np.pi
        * quad(
            lambda x: (
                np.sqrt(1 - x**2)
                * np.exp(-2 * effective_attenuation_factor * np.sqrt(1 - x**2))
            ),
            -1,
            1,
        )[0]
    )


@pytest.mark.parametrize('scattering_cross_section', [0.1, 0.5, 1.0])
def test_compute_transmission_map(scattering_cross_section):
    material = Material(
        ScatteringParams(
            'Fake',
            absorption_cross_section=sc.scalar(0.0, unit='mm**2'),
            total_scattering_cross_section=sc.scalar(
                scattering_cross_section, unit='mm**2'
            ),
        ),
        sc.scalar(1.0, unit='1/mm**3'),
    )
    cylinder = Cylinder(
        sc.vector([0, 1.0, 0]),
        sc.vector([0, 0, 0.0], unit='mm'),
        sc.scalar(1.0, unit='mm'),
        sc.scalar(1.0, unit='mm'),
    )

    tm = compute_transmission_map(
        cylinder,
        material,
        beam_direction=sc.vector([0, 0, 1]),
        wavelength=sc.linspace('wavelength', 1.0, 1, 1, unit='angstrom'),
        detector_position=sc.vectors(dims='x', values=[[0, 0, 1]], unit='m'),
        quadrature_kind='expensive',
    )
    assert_allclose(
        tm['wavelength', 0]['x', 0].data,
        sc.scalar(transmission_fraction_case1(scattering_cross_section)),
        rtol=sc.scalar(1e-3),
    )


@pytest.mark.parametrize('scattering_cross_section', [0.1, 0.5, 1.0])
def test_compute_transmission_map_wavelength_dependent(scattering_cross_section):
    material = Material(
        ScatteringParams(
            'Fake',
            # Modify the absorption cross section
            absorption_cross_section=sc.scalar(
                scattering_cross_section / 5, unit='mm**2'
            ),
            total_scattering_cross_section=sc.scalar(0.0, unit='mm**2'),
        ),
        sc.scalar(1.0, unit='1/mm**3'),
    )
    cylinder = Cylinder(
        sc.vector([0, 1.0, 0]),
        sc.vector([0, 0, 0.0], unit='mm'),
        sc.scalar(1.0, unit='mm'),
        sc.scalar(1.0, unit='mm'),
    )

    tm = compute_transmission_map(
        cylinder,
        material,
        beam_direction=sc.vector([0, 0, 1]),
        wavelength=sc.linspace(
            # Set wavelength so that it cancels the
            # modified absorption cross section
            'wavelength',
            5 * 1.7982,
            5 * 1.7982,
            1,
            unit='angstrom',
        ),
        detector_position=sc.vectors(dims='x', values=[[0, 0, 1]], unit='m'),
        quadrature_kind='expensive',
    )
    assert_allclose(
        tm['wavelength', 0]['x', 0].data,
        sc.scalar(transmission_fraction_case1(scattering_cross_section)),
        rtol=sc.scalar(1e-3),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import scipp as sc


def chopper_mockup() -> sc.DataGroup:
    """Return a data group for a fake chopper.

    Returns
    -------
    :
        Data that looks similar to a chopper that was loaded
        from a NeXus ``NXdisk_chopper``.
    """
    rotation_speed = _rotation_speed()
    return sc.DataGroup(
        {
            'delay': sc.DataGroup(
                {
                    'value': sc.DataArray(
                        sc.array(dims=['time'], values=[3050], unit='ns'),
                        coords={
                            'time': sc.datetimes(
                                dims=['time'], values=['2023-01-19T08:11:06'], unit='ns'
                            )
                        },
                    ),
                }
            ),
            'position': sc.vector([0.0, 0.0, 60.0], unit='m'),
            'radius': sc.scalar(0.35, unit='m'),
            'rotation_speed': sc.DataGroup(
                {
                    'value': rotation_speed,
                }
            ),
            'slit_height': sc.scalar(0.1, unit='m'),
            'slit_edges': sc.array(
                dims=['slit'], values=[30.0, 160.0, 210.0, 280.0], unit='deg'
            ),
            'slits': 2,
            'top_dead_center': sc.DataGroup(
                {
                    'time': _top_dead_center(rotation_speed),
                }
            ),
        }
    )


def _rotation_speed(target_frequency: float = 14.0) -> sc.DataArray:
    time = _time_coord()

    # Arbitrary time scale.
    t = sc.linspace('time', 0, 96, 100, unit='s')
    omega = sc.linspace('time', 0.8, 0.0, 100, unit='rad/s')
    lam = sc.scalar(0.3, unit='1/s')
    phi = sc.scalar(0.9, unit='rad')
    scale = sc.scalar(0.1)
    f_plateau = target_frequency + scale * sc.sin(omega * t + phi) * sc.exp(-lam * t)
    f_plateau.unit = 'Hz'

    shoulder = (100 - np.tanh(0.7 * (t.values - 80))) / 101
    f_plateau.values *= shoulder

    f_rising = sc.linspace('time', 0.0, target_frequency, 40, unit='Hz')
    f_falling = sc.linspace('time', f_plateau[-1].value, 0.0, 40, unit='Hz')
    f = sc.concat([f_rising, f_plateau, f_falling], dim='time')

    rng = np.random.default_rng(84391)
    noise = sc.array(dims=['time'], values=rng.normal(0.0, 4e-5, 180), unit='Hz')

    return sc.DataArray(f + noise, coords={'time': time})


def _top_dead_center(rotation_speed: sc.DataArray) -> sc.Variable:
    """Compute TDC timestamps for a given rotation speed measurement.

    This assumes that the rotation speed was measured at a coarser rate than TDC and
    that the reported speed values are the speed at that moment.
    The function uses a linear interpolation between speed measurements to determine
    the momentary rotation speed.
    """
    # This function first picks a time value τ near the first measured
    # point of the rotation speed.
    # Then, τ is stepped forwards according to the estimated rotation speed at τ,
    # see the docstring about interpolation.
    #
    # Given two neighboring TDC timestamps τ_{i-1} < τ_i, we have
    #     τ_i
    #      ∫ f(t) dt = 1
    #   τ_{i-1}
    # where f(t) is the rotation speed obtained by interpolation.
    # Solving this equation gives a function τ_i(τ_{i-1}) for computing a timestamp
    # τ_i from the previous τ_{i-1}.

    time = rotation_speed.coords['time'] - sc.epoch(
        unit=rotation_speed.coords['time'].unit
    )
    slope, offset = _interpolate_rotation_speed(speed=rotation_speed.data, time=time)

    tdc = []
    # Arbitrary shift to mis-align tdc and speed timestamps.
    tau = time[0].copy().to(dtype='float64') + sc.scalar(12_000_000, unit='ns')
    while tau < time[-1]:
        tdc.append(tau)
        m = slope[tau]
        d = offset[tau]
        if m.value == 0.0:
            tau = tau + sc.reciprocal(d).to(unit=tau.unit)
        else:
            f = sc.to_unit(m * tau, d.unit) + d
            tau = (sc.sqrt(f**2 + 2 * m) - d) / m
            tau = tau.to(unit=time.unit)

    return sc.concat(tdc, dim='time').to(unit='ns', dtype='int64') + sc.epoch(unit='ns')


def _interpolate_rotation_speed(*, speed: sc.Variable, time: sc.Variable):
    """Interpolate the speed-vs-time curve with line segments."""
    m = sc.to_unit((speed[1:] - speed[:-1]) / (time[1:] - time[:-1]), 'Hz^2')
    d = speed[:-1] - sc.to_unit(m * time[:-1], speed.unit)

    slope = sc.lookup(sc.DataArray(m, coords={'time': time.to(dtype='float64')}))
    offset = sc.lookup(sc.DataArray(d, coords={'time': time.to(dtype='float64')}))
    return slope, offset


def _time_coord() -> sc.Variable:
    dt_min = 1674115866205830382
    dt_max = 1674116082246890859
    dt = sc.linspace('time', dt_min, dt_max + 1, 180, dtype='int64', unit='ns')
    return sc.epoch(unit='ns') + dt



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc

import scippneutron as scn


def test_scattering_params_h():
    params = scn.atoms.ScatteringParams.for_isotope('H')
    expected = scn.atoms.ScatteringParams(
        isotope='H',
        coherent_scattering_length_re=sc.scalar(-3.739, unit='fm'),
        coherent_scattering_length_im=None,
        incoherent_scattering_length_re=None,
        incoherent_scattering_length_im=None,
        coherent_scattering_cross_section=sc.scalar(1.7568, unit='barn'),
        incoherent_scattering_cross_section=sc.scalar(80.26, unit='barn'),
        total_scattering_cross_section=sc.scalar(82.02, unit='barn'),
        absorption_cross_section=sc.scalar(0.3326, unit='barn'),
    )
    assert params == expected


def test_scattering_params_157gd():
    params = scn.atoms.ScatteringParams.for_isotope('157Gd')
    expected = scn.atoms.ScatteringParams(
        isotope='157Gd',
        coherent_scattering_length_re=sc.scalar(-1.14, unit='fm'),
        coherent_scattering_length_im=sc.scalar(-71.9, unit='fm'),
        incoherent_scattering_length_re=sc.scalar(5.0, variance=5.0**2, unit='fm'),
        incoherent_scattering_length_im=sc.scalar(-55.8, unit='fm'),
        coherent_scattering_cross_section=sc.scalar(
            650.0, variance=4.0**2, unit='barn'
        ),
        incoherent_scattering_cross_section=sc.scalar(
            394.0, variance=7.0**2, unit='barn'
        ),
        total_scattering_cross_section=sc.scalar(1044.0, variance=8.0**2, unit='barn'),
        absorption_cross_section=sc.scalar(259000.0, variance=700.0**2, unit='barn'),
    )
    assert params == expected


def test_scattering_params_unknown():
    with pytest.raises(ValueError, match="No entry for element / isotope 'scippium'"):
        scn.atoms.ScatteringParams.for_isotope('scippium')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import io
import re
from datetime import datetime, timezone
from pathlib import Path

import pytest
import scipp as sc

from scippneutron import __version__
from scippneutron.io import cif


def write_to_str(block: cif.Block) -> str:
    buffer = io.StringIO()
    block.write(buffer)
    buffer.seek(0)
    return buffer.getvalue()


def save_to_str(cif_: cif.CIF) -> str:
    buffer = io.StringIO()
    cif_.save(buffer)
    buffer.seek(0)
    return buffer.getvalue()


def test_write_block_empty():
    block = cif.Block('a-block-name')
    res = write_to_str(block)
    assert res == 'data_a-block-name\n\n'


def test_write_block_name_with_space():
    with pytest.raises(ValueError, match='Block name must not contain spaces'):
        cif.Block('a block-name with space')


def test_write_block_comment():
    block = cif.Block('a-block-name', comment='some comment\n to describe the block')
    res = write_to_str(block)
    assert (
        res
        == '''# some comment
#  to describe the block
data_a-block-name

'''
    )


def test_write_block_single_pair_string():
    block = cif.Block('single', [{'audit.creation_method': 'written_by_scippneutron'}])
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method written_by_scippneutron
'''
    )


def test_write_block_single_pair_string_variable():
    block = cif.Block(
        'single', [{'audit.creation_method': sc.scalar('written_by_scippneutron')}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method written_by_scippneutron
'''
    )


def test_write_block_single_pair_number():
    block = cif.Block('number', [{'cell.angle_alpha': 62}])
    res = write_to_str(block)
    assert (
        res
        == '''data_number

_cell.angle_alpha 62
'''
    )


@pytest.mark.parametrize('unit', [None, 'deg'])
def test_write_block_single_pair_number_variable(unit):
    block = cif.Block('number', [{'cell.angle_alpha': sc.scalar(93, unit=unit)}])
    res = write_to_str(block)
    assert (
        res
        == '''data_number

_cell.angle_alpha 93
'''
    )


@pytest.mark.parametrize('unit', [None, 'deg'])
def test_write_block_single_pair_number_error(unit):
    block = cif.Block(
        'number', [{'cell.angle_alpha': sc.scalar(93.2, variance=2.1**2, unit=unit)}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_number

_cell.angle_alpha 93(2)
'''
    )


def test_write_block_single_pair_datetime():
    dt = datetime(
        year=2023, month=12, day=1, hour=15, minute=9, second=45, tzinfo=timezone.utc
    )
    block = cif.Block(
        'datetime',
        [
            {
                'audit.creation_date': dt,
            }
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_datetime

_audit.creation_date 2023-12-01T15:09:45+00:00
'''
    )


def test_write_block_single_pair_datetime_variable():
    block = cif.Block(
        'datetime',
        [
            {
                'audit.creation_date': sc.datetime('2023-12-01T15:12:33'),
            }
        ],
    )
    res = write_to_str(block)
    # No timezone info in the output!
    assert (
        res
        == '''data_datetime

_audit.creation_date 2023-12-01T15:12:33
'''
    )


def test_write_block_single_pair_space():
    block = cif.Block('single', [{'audit.creation_method': 'written by scippneutron'}])
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method 'written by scippneutron'
'''
    )


def test_write_block_single_pair_single_quote():
    block = cif.Block(
        'single', [{'audit.creation_method': "written by 'scippneutron'"}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method "written by 'scippneutron'"
'''
    )


def test_write_block_single_pair_double_quote():
    block = cif.Block(
        'single', [{'audit.creation_method': 'written by "scippneutron"'}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method 'written by "scippneutron"'
'''
    )


def test_write_block_single_pair_both_quotes():
    block = cif.Block(
        'single', [{'audit.creation_method': """'written by "scippneutron"'"""}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method
; 'written by "scippneutron"'
;
'''
    )


def test_write_block_single_pair_newline():
    block = cif.Block(
        'single',
        [{'audit.creation_method': "written by scippneutron\n    version 2000"}],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method
; written by scippneutron
    version 2000
;
'''
    )


def test_write_block_single_pair_utf8():
    block = cif.Block('utf-8', [{'audit.creation_method': 'Unicode: \xb5\xc5'}])
    res = write_to_str(block)
    assert (
        res
        == r'''data_utf-8

_audit.creation_method 'Unicode: \xb5\xc5'
'''
    )


def test_write_block_single_pair_empty_value():
    block = cif.Block('utf-8', [{'audit.creation_method': ''}])
    res = write_to_str(block)
    assert (
        res
        == r'''data_utf-8

_audit.creation_method ''
'''
    )


def test_write_block_single_pair_single_line_comment():
    block = cif.Block('comment')
    block.add({'diffrn_radiation.probe': 'neutron'}, comment='a comment')
    res = write_to_str(block)
    assert (
        res
        == '''data_comment

# a comment
_diffrn_radiation.probe neutron
'''
    )


def test_write_block_single_pair_multi_line_comment():
    block = cif.Block('comment')
    block.add(
        {'diffrn_radiation.probe': 'neutron'},
        comment='Guessing that\nthis is the\n  correct probe',
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_comment

# Guessing that
# this is the
#   correct probe
_diffrn_radiation.probe neutron
'''
    )


def test_write_block_single_pair_single_line_comment_utf8():
    block = cif.Block('comment')
    block.add({'diffrn_radiation.probe': 'neutron'}, comment='unicode: \xc5')
    res = write_to_str(block)
    assert (
        res
        == r'''data_comment

# unicode: \xc5
_diffrn_radiation.probe neutron
'''
    )


def test_write_block_multiple_pairs():
    block = cif.Block(
        'multiple',
        [
            {
                'audit.creation_method': 'written_by_scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            }
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_multiple

_audit.creation_method written_by_scippneutron
_audit.creation_date 2023-12-01T13:52:00Z
'''
    )


def test_write_block_multiple_chunks():
    block = cif.Block(
        'multiple',
        [
            {
                'audit.creation_method': 'written_by_scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            }
        ],
    )
    block.add({'diffrn_radiation.probe': 'neutron'})
    res = write_to_str(block)
    assert (
        res
        == '''data_multiple

_audit.creation_method written_by_scippneutron
_audit.creation_date 2023-12-01T13:52:00Z

_diffrn_radiation.probe neutron
'''
    )


def test_write_block_multiple_chunks_comment():
    block = cif.Block(
        'multiple',
        [
            {
                'audit.creation_method': 'written_by_scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            }
        ],
    )
    block.add({'diffrn_radiation.probe': 'neutron'}, comment='Guessed')
    res = write_to_str(block)
    assert (
        res
        == '''data_multiple

_audit.creation_method written_by_scippneutron
_audit.creation_date 2023-12-01T13:52:00Z

# Guessed
_diffrn_radiation.probe neutron
'''
    )


def test_write_block_single_loop_one_column():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block('looped', [cif.Loop({'diffrn.ambient_environment': env})])
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
water
sulfur
'''
    )


def test_write_block_single_loop_one_column_comment():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block(
        'looped',
        [
            cif.Loop(
                {'diffrn.ambient_environment': env},
                comment='This data is completely made up!',
            )
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

# This data is completely made up!
loop_
_diffrn.ambient_environment
water
sulfur
'''
    )


def test_write_block_single_loop_two_columns():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    id_ = sc.array(dims=['x'], values=['123', 'x6a'])
    block = cif.Block(
        'looped', [cif.Loop({'diffrn.ambient_environment': env, 'diffrn.id': id_})]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
_diffrn.id
water 123
sulfur x6a
'''
    )


def test_write_block_single_loop_multi_line_string():
    env = sc.array(dims=['x'], values=['water\nand some salt', 'sulfur'])
    id_ = sc.array(dims=['x'], values=['123', 'x6a'])
    block = cif.Block(
        'looped', [cif.Loop({'diffrn.ambient_environment': env, 'diffrn.id': id_})]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
_diffrn.id
; water
and some salt
;
123
sulfur
x6a
'''
    )


def test_write_block_single_loop_empty_string():
    env = sc.array(dims=['x'], values=['', 'sulfur'])
    id_ = sc.array(dims=['x'], values=['123', 'x6a'])
    block = cif.Block(
        'looped', [cif.Loop({'diffrn.ambient_environment': env, 'diffrn.id': id_})]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
_diffrn.id
'' 123
sulfur x6a
'''
    )


def test_write_block_single_loop_numbers():
    coeff = sc.array(dims=['cal'], values=[3.65, -0.012, 1.2e-5])
    power = sc.array(dims=['cal'], values=[0, 1, 2])
    id_ = sc.array(dims=['cal'], values=['tzero', 'DIFC', 'DIFA'])
    block = cif.Block(
        'looped',
        [
            cif.Loop(
                {
                    'pd_calib_d_to_tof.id': id_,
                    'pd_calib_d_to_tof.power': power,
                    'pd_calib_d_to_tof.coeff': coeff,
                }
            )
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_pd_calib_d_to_tof.id
_pd_calib_d_to_tof.power
_pd_calib_d_to_tof.coeff
tzero 0 3.65
DIFC 1 -0.012
DIFA 2 1.2e-05
'''
    )


def test_write_block_single_loop_numbers_errors():
    coeff = sc.array(
        dims=['cal'],
        values=[3.65, -0.012, 1.2e-5],
        variances=[0.13**2, 0.001**2, 2e-6**2],
    )
    power = sc.array(dims=['cal'], values=[0, 1, 2])
    id_ = sc.array(dims=['cal'], values=['tzero', 'DIFC', 'DIFA'])
    block = cif.Block(
        'looped',
        [
            cif.Loop(
                {
                    'pd_calib_d_to_tof.id': id_,
                    'pd_calib_d_to_tof.power': power,
                    'pd_calib_d_to_tof.coeff': coeff,
                }
            )
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_pd_calib_d_to_tof.id
_pd_calib_d_to_tof.power
_pd_calib_d_to_tof.coeff
tzero 0 3.65(13)
DIFC 1 -0.0120(10)
DIFA 2 0.000012(2)
'''
    )


def test_write_block_pairs_then_loop():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block(
        'looped',
        [
            {
                'audit.creation_method': 'written by scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            },
            cif.Loop({'diffrn.ambient_environment': env}),
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

_audit.creation_method 'written by scippneutron'
_audit.creation_date 2023-12-01T13:52:00Z

loop_
_diffrn.ambient_environment
water
sulfur
'''
    )


def test_write_block_loop_then_pairs():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block(
        'looped',
        [
            cif.Loop({'diffrn.ambient_environment': env}),
            {
                'audit.creation_method': 'written by scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            },
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
water
sulfur

_audit.creation_method 'written by scippneutron'
_audit.creation_date 2023-12-01T13:52:00Z
'''
    )


def test_write_block_pair_then_loop_then_pairs():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block(
        'looped',
        [
            {'diffrn_radiation.probe': 'neutron'},
            cif.Loop({'diffrn.ambient_environment': env}),
            {
                'audit.creation_method': 'written by scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            },
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

_diffrn_radiation.probe neutron

loop_
_diffrn.ambient_environment
water
sulfur

_audit.creation_method 'written by scippneutron'
_audit.creation_date 2023-12-01T13:52:00Z
'''
    )


def test_write_block_two_loops():
    env = sc.array(dims=['env'], values=['water', 'sulfur'])
    author = sc.array(dims=['author'], values=['Ridcully, M.', 'Librarian'])
    email = sc.array(dims=['author'], values=['m.ridcully@uu.am', 'lib@uu.am'])
    block = cif.Block(
        'looped',
        [
            cif.Loop({'diffrn.ambient_environment': env}),
            cif.Loop({'audit_author.name': author, 'audit_author.email': email}),
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
water
sulfur

loop_
_audit_author.name
_audit_author.email
'Ridcully, M.' m.ridcully@uu.am
Librarian lib@uu.am
'''
    )


def test_write_block_core_schema_from_chunk():
    chunk = cif.Chunk(
        {'audit.creation_method': 'written by scippneutron'}, schema=cif.CORE_SCHEMA
    )
    block = cif.Block('block-with-schema', [chunk])
    res = write_to_str(block)
    assert (
        res
        == '''data_block-with-schema

loop_
_audit_conform.dict_name
_audit_conform.dict_version
_audit_conform.dict_location
coreCIF 3.3.0 https://github.com/COMCIFS/cif_core/blob/6f8502e81b623eb0fd779c79efaf191d49fa198c/cif_core.dic

_audit.creation_method 'written by scippneutron'
'''
    )


def test_write_block_core_schema_from_loop():
    author = sc.array(dims=['author'], values=['Ridcully, M.', 'Librarian'])
    email = sc.array(dims=['author'], values=['m.ridcully@uu.am', 'lib@uu.am'])
    loop = cif.Loop(
        {'audit_author.name': author, 'audit_author.email': email},
        schema=cif.CORE_SCHEMA,
    )
    block = cif.Block('block-with-schema', [loop])
    res = write_to_str(block)
    assert (
        res
        == '''data_block-with-schema

loop_
_audit_conform.dict_name
_audit_conform.dict_version
_audit_conform.dict_location
coreCIF 3.3.0 https://github.com/COMCIFS/cif_core/blob/6f8502e81b623eb0fd779c79efaf191d49fa198c/cif_core.dic

loop_
_audit_author.name
_audit_author.email
'Ridcully, M.' m.ridcully@uu.am
Librarian lib@uu.am
'''
    )


def test_write_block_pd_schema_from_chunk():
    chunk = cif.Chunk(
        {'pd_meas.units_of_intensity': '1/(micro ampere)'}, schema=cif.PD_SCHEMA
    )
    block = cif.Block('block-with-schema', [chunk])
    res = write_to_str(block)
    # The order of schemas is arbitrary, so we cannot easily check the whole string.
    assert 'pdCIF' in res
    assert 'coreCIF' in res


def test_write_block_multi_schema_schema():
    core_chunk = cif.Chunk(
        {'audit.creation_method': 'written by scippneutron'}, schema=cif.CORE_SCHEMA
    )
    pd_chunk = cif.Chunk(
        {'pd_meas.units_of_intensity': '1/(micro ampere)'}, schema=cif.PD_SCHEMA
    )
    block = cif.Block('block-with-schema', [core_chunk, pd_chunk])
    res = write_to_str(block)
    # The order of schemas is arbitrary, so we cannot easily check the whole string.
    assert 'pdCIF' in res
    assert 'coreCIF' in res


def test_save_cif_one_block_buffer():
    block1 = cif.Block(
        'block-1', [{'audit.creation_method': 'written by scippneutron'}]
    )
    buffer = io.StringIO()
    cif.save_cif(buffer, block1)
    buffer.seek(0)
    assert (
        buffer.read()
        == r'''#\#CIF_1.1
data_block-1

_audit.creation_method 'written by scippneutron'
'''
    )


def test_save_cif_two_blocks_buffer():
    env = sc.array(dims=['env'], values=['water', 'sulfur'])
    block1 = cif.Block(
        'block-1', [{'audit.creation_method': 'written by scippneutron'}]
    )
    block2 = cif.Block(
        'block-2',
        [
            {'diffrn_radiation.probe': 'neutron'},
            cif.Loop({'diffrn.ambient_environment': env}),
        ],
    )
    buffer = io.StringIO()
    cif.save_cif(buffer, [block1, block2])
    buffer.seek(0)
    assert (
        buffer.read()
        == r'''#\#CIF_1.1
data_block-1

_audit.creation_method 'written by scippneutron'

data_block-2

_diffrn_radiation.probe neutron

loop_
_diffrn.ambient_environment
water
sulfur
'''
    )


@pytest.mark.parametrize('path_type', [str, Path])
def test_save_cif_one_block_file(tmpdir, path_type):
    path = path_type(Path(tmpdir) / "test_save_cif_one_block.cif")
    block1 = cif.Block(
        'block-1', [{'audit.creation_method': 'written by scippneutron'}]
    )

    cif.save_cif(path, block1)
    with open(path) as f:
        assert (
            f.read()
            == r'''#\#CIF_1.1
data_block-1

_audit.creation_method 'written by scippneutron'
'''
        )


def test_loop_requires_1d():
    with pytest.raises(sc.DimensionError):
        cif.Loop({'fake': sc.zeros(sizes={'x': 4, 'y': 3})})


def test_loop_requires_matching_dims():
    with pytest.raises(sc.DimensionError):
        cif.Loop({'a': sc.zeros(sizes={'x': 4}), 'b': sc.zeros(sizes={'x': 3})})
    with pytest.raises(sc.DimensionError):
        cif.Loop({'a': sc.zeros(sizes={'x': 4}), 'b': sc.zeros(sizes={'y': 4})})


def test_builder_writes_comment() -> None:
    cif_ = cif.CIF(comment='This is a test comment\nacross multiple lines.')
    res = save_to_str(cif_)

    expected_start = r'''#\#CIF_1.1
# This is a test comment
# across multiple lines.
data_
'''
    assert res.startswith(expected_start)


def test_builder_writes_audit() -> None:
    cif_ = cif.CIF()
    res = save_to_str(cif_)

    # Escape + to make this usable in a regex
    expected_version = str(__version__).replace('+', r'\+')
    expected = re.compile(rf'''#\\#CIF_1.1
data_

loop_
_audit_conform.dict_name
_audit_conform.dict_version
_audit_conform.dict_location
coreCIF 3.3.0 https://github.com/COMCIFS/cif_core/blob/6f8502e81b623eb0fd779c79efaf191d49fa198c/cif_core.dic

_audit.creation_date \d{{4}}-\d{{2}}-\d{{2}}T\d{{2}}:\d{{2}}:\d{{2}}\+00.00
_audit.creation_method 'Written by scippneutron v{expected_version}'
''')
    assert re.match(expected, res)


def test_builder_writes_audit_with_one_reducer() -> None:
    cif_ = cif.CIF().with_reducers('mypackage vFINAL')
    res = save_to_str(cif_)

    # Escape + to make this usable in a regex
    expected_version = str(__version__).replace('+', r'\+')
    expected = re.compile(rf'''#\\#CIF_1.1
data_

loop_
_audit_conform.dict_name
_audit_conform.dict_version
_audit_conform.dict_location
coreCIF 3.3.0 https://github.com/COMCIFS/cif_core/blob/6f8502e81b623eb0fd779c79efaf191d49fa198c/cif_core.dic

_audit.creation_date \d{{4}}-\d{{2}}-\d{{2}}T\d{{2}}:\d{{2}}:\d{{2}}\+00.00
_audit.creation_method 'Written by scippneutron v{expected_version}'
_computing.diffrn_reduction 'mypackage vFINAL'
''')
    assert re.match(expected, res)


def test_builder_writes_audit_with_multiple_reducers() -> None:
    cif_ = cif.CIF().with_reducers('package 1', 'package 2')
    res = save_to_str(cif_)

    # Escape + to make this usable in a regex
    expected_version = str(__version__).replace('+', r'\+')
    expected = re.compile(rf'''#\\#CIF_1.1
data_

loop_
_audit_conform.dict_name
_audit_conform.dict_version
_audit_conform.dict_location
coreCIF 3.3.0 https://github.com/COMCIFS/cif_core/blob/6f8502e81b623eb0fd779c79efaf191d49fa198c/cif_core.dic

_audit.creation_date \d{{4}}-\d{{2}}-\d{{2}}T\d{{2}}:\d{{2}}:\d{{2}}\+00.00
_audit.creation_method 'Written by scippneutron v{expected_version}'

loop_
_computing.diffrn_reduction
'package 1'
'package 2'
''')
    assert re.match(expected, res)


def test_builder_with_beamline() -> None:
    original = cif.CIF()
    cif_ = original.with_beamline(beamline='DREAM', facility='ESS')
    res = save_to_str(cif_)
    original_res = save_to_str(original)

    expected = '''_diffrn_radiation.probe neutron
_diffrn_source.beamline DREAM
_diffrn_source.facility ESS
_diffrn_source.device spallation'''
    assert expected in res
    assert expected not in original_res


def test_builder_with_reduced_powder_data():
    da = sc.DataArray(
        sc.array(
            dims=['tof'],
            values=[13.6, 26.0, 9.7],
            variances=[0.81, 1, 0.36],
        ),
        coords={'tof': sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='us')},
    )

    cif_ = cif.CIF('reduced')
    cif_ = cif_.with_reduced_powder_data(da)
    res = save_to_str(cif_)

    assert 'pdCIF' in res
    assert 'coreCIF' in res

    _, tof_loop = res.rsplit('\n\n', 1)
    assert (
        tof_loop
        == '''loop_
_pd_data.point_id
_pd_meas.time_of_flight
_pd_proc.intensity_norm
_pd_proc.intensity_norm_su
0 1.2 13.6 0.9
1 1.4 26.0 1.0
2 2.3 9.7 0.6
'''
    )


def test_builder_with_reduced_powder_data_custom_unit():
    da = sc.DataArray(
        sc.array(dims=['tof'], values=[13.6, 26.0, 9.7], unit='counts'),
        coords={'tof': sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='us')},
    )

    cif_ = cif.CIF('reduced')
    cif_ = cif_.with_reduced_powder_data(da)
    res = save_to_str(cif_)

    assert 'pdCIF' in res
    assert 'coreCIF' in res

    _, tof_loop = res.rsplit('\n\n', 1)
    assert (
        tof_loop
        == '''# Unit of intensity: [counts]
loop_
_pd_data.point_id
_pd_meas.time_of_flight
_pd_proc.intensity_norm
0 1.2 13.6
1 1.4 26.0
2 2.3 9.7
'''
    )


def test_builder_with_reduced_powder_data_bad_dim():
    da = sc.DataArray(
        sc.array(
            dims=['time'],
            values=[13.6, 26.0, 9.7],
        ),
        coords={'time': sc.array(dims=['time'], values=[1.2, 1.4, 2.3], unit='us')},
    )

    cif_ = cif.CIF('reduced')
    with pytest.raises(sc.CoordError):
        cif_.with_reduced_powder_data(da)


def test_builder_with_reduced_powder_data_bad_name():
    da = sc.DataArray(
        sc.array(
            dims=['tof'],
            values=[13.6, 26.0, 9.7],
        ),
        coords={'tof': sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='us')},
        name='bad',
    )

    cif_ = cif.CIF('reduced')
    with pytest.raises(
        ValueError, match='Unrecognized name for reduced powder data: bad'
    ):
        cif_.with_reduced_powder_data(da)


def test_builder_with_reduced_powder_data_bad_coord_unit():
    da = sc.DataArray(
        sc.array(dims=['tof'], values=[13.6, 26.0, 9.7]),
        coords={'tof': sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='ns')},
    )

    cif_ = cif.CIF('reduced')
    with pytest.raises(sc.UnitError):
        cif_.with_reduced_powder_data(da)


def test_builder_powder_calibration():
    da = sc.DataArray(
        sc.array(dims=['cal'], values=[1.2, 4.5, 6.7]),
        coords={'power': sc.array(dims=['cal'], values=[0, 1, -1])},
    )
    cif_ = cif.CIF('cal')
    cif_ = cif_.with_powder_calibration(da)
    res = save_to_str(cif_)

    assert 'pdCIF' in res
    assert 'coreCIF' in res

    _, cal_loop = res.rsplit('\n\n', 1)
    assert (
        cal_loop
        == '''loop_
_pd_calib_d_to_tof.id
_pd_calib_d_to_tof.power
_pd_calib_d_to_tof.coeff
ZERO 0 1.2
DIFC 1 4.5
DIFB -1 6.7
'''
    )


def test_builder_single_contact_author() -> None:
    author = cif.Author(
        name='Jane Doe',
        email='jane.doe@ess.eu',
        address='Partikelgatan, Lund',
        orcid='https://orcid.org/0000-0000-0000-0001',
    )
    cif_ = cif.CIF().with_authors(author)
    result = save_to_str(cif_)
    expected = '''_audit_contact_author.name 'Jane Doe'
_audit_contact_author.email jane.doe@ess.eu
_audit_contact_author.address 'Partikelgatan, Lund'
_audit_contact_author.id_orcid https://orcid.org/0000-0000-0000-0001'''
    assert expected in result


def test_builder_regular_author() -> None:
    author = cif.Author(
        name='Jane Doe',
        email='jane.doe@ess.eu',
        address='Partikelgatan, Lund',
        orcid='https://orcid.org/0000-0000-0000-0001',
        corresponding=False,
    )
    cif_ = cif.CIF().with_authors(author)
    result = save_to_str(cif_)
    expected = '''_audit_author.name 'Jane Doe'
_audit_author.email jane.doe@ess.eu
_audit_author.address 'Partikelgatan, Lund'
_audit_author.id_orcid https://orcid.org/0000-0000-0000-0001'''
    assert expected in result


def test_builder_multiple_regular_authors() -> None:
    authors = [
        cif.Author(
            name='Jane Doe',
            email='jane.doe@ess.eu',
            address='Partikelgatan, Lund',
            orcid='https://orcid.org/0000-0000-0000-0001',
            corresponding=False,
        ),
        cif.Author(
            name='Max Mustermann',
            email='mm@scipp.eu',
            orcid='https://orcid.org/0000-0000-0000-0002',
            corresponding=False,
        ),
    ]
    cif_ = cif.CIF().with_authors(*authors)
    result = save_to_str(cif_)
    expected = """loop_
_audit_author.name
_audit_author.email
_audit_author.address
_audit_author.id_orcid
'Jane Doe' jane.doe@ess.eu 'Partikelgatan, Lund' https://orcid.org/0000-0000-0000-0001
'Max Mustermann' mm@scipp.eu '' https://orcid.org/0000-0000-0000-0002
"""
    assert expected in result


def test_builder_regular_author_role() -> None:
    author = cif.Author(
        name='Jane Doe',
        role='measurement',
        corresponding=False,
    )
    cif_ = cif.CIF().with_authors(author)
    result = save_to_str(cif_)

    author_pattern = re.compile(r"""_audit_author.name 'Jane Doe'
_audit_author.id ([0-9a-f]+)""")
    author_match = re.search(author_pattern, result)
    assert author_match is not None
    author_id = author_match.group(1)

    expected = rf"""loop_
_audit_author_role.id
_audit_author_role.role
{author_id} measurement"""

    assert expected in result


def test_builder_many_fields() -> None:
    cif_ = (
        cif.CIF('my/name')
        .with_authors(cif.Author(name='Jane Doe'))
        .with_reducers('test-package')
        .with_beamline(beamline='fake')
    )
    result = save_to_str(cif_)

    assert 'data_my/name' in result
    assert '_audit_contact_author.name' in result
    assert '_diffrn_radiation.probe neutron' in result
    assert '_diffrn_source.beamline fake' in result
    assert '_computing.diffrn_reduction test-package'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc

from scippneutron.chopper import DiskChopperType, extract_chopper_from_nexus


@pytest.fixture
def raw_nexus_chopper():
    return sc.DataGroup(
        {
            'type': 'Chopper type single',
            'position': sc.vector([0.0, 0.0, 2.0], unit='m'),
            'rotation_speed': sc.scalar(12.0, unit='Hz'),
            'beam_position': sc.scalar(45.0, unit='deg'),
            'phase': sc.scalar(-20.0, unit='deg'),
            'slit_edges': sc.array(
                dims=['dim_0'], values=[0.0, 60.0, 124.0, 126.0], unit='deg'
            ),
            'slit_height': sc.array(dims=['slit'], values=[0.4, 0.3], unit='m'),
            'radius': sc.scalar(0.5, unit='m'),
            'top_dead_center': sc.datetimes(
                dims=['time'], values=[12, 56, 78], unit='ms'
            ),
        }
    )


def test_post_process_assigns_default_type(raw_nexus_chopper):
    del raw_nexus_chopper['type']
    processed = extract_chopper_from_nexus(raw_nexus_chopper)
    assert processed['type'] == DiskChopperType.single


def test_post_process_extracts_from_logs(raw_nexus_chopper):
    raw_nexus_chopper['rotation_speed'] = sc.DataGroup(
        {
            'value': sc.array(dims=['time'], values=[12, 56, 78], unit='Hz'),
            'time': sc.datetimes(dims=['time'], values=[2, 5, 8], unit='s'),
        }
    )
    processed = extract_chopper_from_nexus(raw_nexus_chopper)
    assert sc.identical(
        processed['rotation_speed'],
        sc.array(dims=['time'], values=[12, 56, 78], unit='Hz'),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock, Neil Vaytet

import os
import re
import uuid
import warnings
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path
from typing import Any

import numpy as np
import scipp as sc
from scipp.core.util import VisibleDeprecationWarning

from ._utils import get_attrs


@contextmanager
def run_mantid_alg(alg, *args, **kwargs):
    try:
        from mantid import simpleapi as mantid
        from mantid.api import AnalysisDataService
    except ImportError:
        raise ImportError(
            "Mantid Python API was not found, please install Mantid framework "
            "as detailed in the installation instructions ("
            "https://scipp.github.io/scippneutron/getting-started/"
            "installation.html)"
        ) from None
    # Deal with multiple calls to this function, which may have conflicting
    # names in the global AnalysisDataService by using uuid.
    ws_name = f'scipp.run_mantid_alg.{uuid.uuid4()}'
    # Deal with non-standard ways to define the prefix of output workspaces
    if alg == 'Fit':
        kwargs['Output'] = ws_name
    elif alg == 'LoadDiffCal':
        kwargs['WorkspaceName'] = ws_name
    else:
        kwargs['OutputWorkspace'] = ws_name
    ws = getattr(mantid, alg)(*args, **kwargs)
    try:
        yield ws
    finally:
        for name in AnalysisDataService.Instance().getObjectNames():
            if name.startswith(ws_name):
                try:
                    mantid.DeleteWorkspace(name)
                except ValueError:
                    # The workspace was deleted already,
                    pass


def get_pos(pos):
    return (
        None
        if pos is None
        else sc.vector(value=[pos.X(), pos.Y(), pos.Z()], unit=sc.units.m)
    )


def make_run(ws):
    return sc.scalar(deepcopy(ws.run()))


additional_unit_mapping = {
    " ": sc.units.one,
    "none": sc.units.one,
}


def process_run_logs(ws):
    for property_name in ws.run().keys():
        units_string = ws.run()[property_name].units
        try:
            unit = additional_unit_mapping.get(units_string, sc.Unit(units_string))
        except RuntimeError:  # TODO catch UnitError once exposed from C++
            # Parsing unit string failed
            unit = None

        values = deepcopy(ws.run()[property_name].value)

        if units_string and unit is None:
            warnings.warn(
                f"Workspace run log '{property_name}' "
                f"has unrecognised units: '{units_string}'",
                stacklevel=4,
            )
        if unit is None:
            unit = sc.units.one

        try:
            times = deepcopy(ws.run()[property_name].times)
            is_time_series = True
            dimension_label = "time"
        except AttributeError:
            times = None
            is_time_series = False
            dimension_label = property_name

        if np.isscalar(values):
            property_data = sc.scalar(values, unit=unit)
        else:
            property_data = sc.array(values=values, unit=unit, dims=[dimension_label])

        if is_time_series:
            # If property has timestamps, create a DataArray
            data_array = sc.DataArray(
                data=property_data,
                coords={
                    dimension_label: sc.array(dims=[dimension_label], values=times)
                },
            )
            yield property_name, data_array
        else:
            yield property_name, property_data


def make_mantid_sample(ws):
    return sc.scalar(deepcopy(ws.sample()))


def make_sample_ub(ws):
    # B matrix transforms the h,k,l triplet into a Cartesian system
    # https://docs.mantidproject.org/nightly/concepts/Lattice.html
    return sc.spatial.linear_transform(
        value=ws.sample().getOrientedLattice().getUB(), unit=sc.units.angstrom**-1
    )


def make_sample_u(ws):
    # U matrix rotation for sample alignment
    # https://docs.mantidproject.org/nightly/concepts/Lattice.html
    return sc.spatial.linear_transform(value=ws.sample().getOrientedLattice().getU())


def make_component_info(ws):
    component_info = ws.componentInfo()

    if component_info.hasSource():
        sourcePos = component_info.sourcePosition()
    else:
        sourcePos = None

    if component_info.hasSample():
        samplePos = component_info.samplePosition()
    else:
        samplePos = None

    return get_pos(sourcePos), get_pos(samplePos)


def make_detector_info(ws, spectrum_dim):
    det_info = ws.detectorInfo()
    # det -> spec mapping
    nDet = det_info.size()
    spectrum = np.empty(shape=(nDet,), dtype=np.int32)
    has_spectrum = np.full((nDet,), False)
    spec_info = ws.spectrumInfo()
    for i, spec in enumerate(spec_info):
        spec_def = spec.spectrumDefinition
        for j in range(len(spec_def)):
            det, time = spec_def[j]
            if time != 0:
                raise RuntimeError(
                    "Conversion of Mantid Workspace with scanning instrument "
                    "not supported yet."
                )
            spectrum[det] = i
            has_spectrum[det] = True

    # Store only information about detectors with data (a spectrum). The rest
    # mostly just gets in the way and including it in the default converter
    # is probably not required.
    spectrum = sc.array(dims=['detector'], values=spectrum[has_spectrum], unit=None)
    detector = sc.array(
        dims=['detector'], values=det_info.detectorIDs()[has_spectrum], unit=None
    )

    # May want to include more information here, such as detector positions,
    # but for now this is not necessary.

    return sc.Dataset(coords={'detector': detector, spectrum_dim: spectrum})


def md_dimension(mantid_dim, index):
    # Look for q dimensions
    patterns = [f"^q.*{coord}$" for coord in ['x', 'y', 'z']]
    q_dims = ['Q_x', 'Q_y', 'Q_z']
    pattern_result = zip(patterns, q_dims, strict=True)
    if mantid_dim.getMDFrame().isQ():
        for pattern, result in pattern_result:
            if re.search(pattern, mantid_dim.name, re.IGNORECASE):
                return result

    # Look for common/known mantid dimensions
    patterns = ["DeltaE", "T"]
    dims = ['energy_transfer', 'temperature']
    pattern_result = zip(patterns, dims, strict=True)
    for pattern, result in pattern_result:
        if re.search(pattern, mantid_dim.name, re.IGNORECASE):
            return result

    # Look for common spatial dimensions
    patterns = [f"^{coord}$" for coord in ['x', 'y', 'z']]
    dims = ['x', 'y', 'z']
    pattern_result = zip(patterns, dims, strict=True)
    for pattern, result in pattern_result:
        if re.search(pattern, mantid_dim.name, re.IGNORECASE):
            return result

    raise ValueError(
        f"Cannot infer scipp dimension from input mantid dimension {mantid_dim.name()}"
    )


def md_unit(frame):
    known_md_units = {
        "Angstrom^-1": sc.units.dimensionless / sc.units.angstrom,
        "r.l.u": sc.units.dimensionless,
        "T": sc.units.K,
        "DeltaE": sc.units.meV,
    }
    if frame.getUnitLabel().ascii() in known_md_units:
        return known_md_units[frame.getUnitLabel().ascii()]
    else:
        return sc.units.dimensionless


def validate_and_get_unit(unit, allow_empty=False):
    if hasattr(unit, 'unitID'):
        if unit.unitID() == 'Label':
            unit = unit.caption()
        else:
            unit = unit.unitID()
    known_units = {
        "DeltaE": ['energy_transfer', sc.units.meV],
        "TOF": ['tof', sc.units.us],
        "Wavelength": ['wavelength', sc.units.angstrom],
        "Energy": ['energy', sc.units.meV],
        "dSpacing": ['dspacing', sc.units.angstrom],
        "MomentumTransfer": ['Q', sc.units.dimensionless / sc.units.angstrom],
        "QSquared": [
            'Q^2',
            sc.units.dimensionless / (sc.units.angstrom * sc.units.angstrom),
        ],
        "Spectrum": ['spectrum', None],
        "Empty": ['empty', sc.units.dimensionless],
        "Counts": ['counts', sc.units.counts],
    }

    if unit not in known_units.keys():
        return [str(unit), sc.units.dimensionless]
    else:
        return known_units[unit]


def _to_spherical(pos, output):
    output["r"] = sc.sqrt(sc.dot(pos, pos))
    output["t"] = sc.acos(pos.fields.z / output["r"].data)
    signed_phi = sc.atan2(y=pos.fields.y, x=pos.fields.x)
    abs_phi = sc.abs(signed_phi)
    output["p-delta"] = (
        np.pi * sc.units.rad
    ) - abs_phi  # angular delta (magnitude) from pole
    output['p-sign'] = signed_phi  # weighted sign of phi
    return output


def _rot_from_vectors(vec1, vec2):
    a = sc.vector(value=vec1.value / np.linalg.norm(vec1.value))
    b = sc.vector(value=vec2.value / np.linalg.norm(vec2.value))
    c = sc.vector(value=np.cross(a.value, b.value))
    angle = sc.acos(sc.dot(a, b)).value
    return sc.spatial.rotation(
        value=[*(c.value * np.sin(angle / 2)), np.cos(angle / 2)]
    )


def get_detector_pos(ws, spectrum_dim):
    nHist = ws.getNumberHistograms()
    pos = np.zeros([nHist, 3])

    spec_info = ws.spectrumInfo()
    for i in range(nHist):
        if spec_info.hasDetectors(i):
            p = spec_info.position(i)
            pos[i, 0] = p.X()
            pos[i, 1] = p.Y()
            pos[i, 2] = p.Z()
        else:
            pos[i, :] = [np.nan, np.nan, np.nan]
    return sc.vectors(dims=[spectrum_dim], values=pos, unit=sc.units.m)


def get_detector_properties(
    ws, source_pos, sample_pos, spectrum_dim, advanced_geometry=False
):
    if not advanced_geometry:
        return (get_detector_pos(ws, spectrum_dim), None, None)
    spec_info = ws.spectrumInfo()
    det_info = ws.detectorInfo()
    comp_info = ws.componentInfo()
    nspec = len(spec_info)
    det_rot_quaternions = np.zeros([nspec, 4])
    det_bbox = np.zeros([nspec, 3])

    if sample_pos is not None and source_pos is not None:
        total_detectors = spec_info.detectorCount()
        act_beam = sample_pos - source_pos
        rot = _rot_from_vectors(act_beam, sc.vector(value=[0, 0, 1]))
        inv_rot = _rot_from_vectors(sc.vector(value=[0, 0, 1]), act_beam)

        # Create empty to hold position info for all spectra detectors
        pos_d = sc.Dataset(
            {"x": sc.zeros(dims=["detector"], shape=[total_detectors], unit=sc.units.m)}
        )
        pos_d["y"] = sc.zeros_like(pos_d["x"])
        pos_d["z"] = sc.zeros_like(pos_d["x"])
        pos_d.coords[spectrum_dim] = sc.array(
            dims=["detector"], values=np.empty(total_detectors)
        )

        spectrum_values = pos_d.coords[spectrum_dim].values

        x_values = pos_d["x"].values
        y_values = pos_d["y"].values
        z_values = pos_d["z"].values

        idx = 0
        for i, spec in enumerate(spec_info):
            if spec.hasDetectors:
                definition = spec_info.getSpectrumDefinition(i)
                n_dets = len(definition)
                quats = []
                bboxes = []
                for j in range(n_dets):
                    det_idx = definition[j][0]
                    p = det_info.position(det_idx)
                    r = det_info.rotation(det_idx)
                    spectrum_values[idx] = i
                    x_values[idx] = p.X()
                    y_values[idx] = p.Y()
                    z_values[idx] = p.Z()
                    idx += 1
                    quats.append(np.array([r.imagI(), r.imagJ(), r.imagK(), r.real()]))
                    if comp_info.hasValidShape(det_idx):
                        s = comp_info.shape(det_idx)
                        bboxes.append(s.getBoundingBox().width())
                det_rot_quaternions[i] = np.mean(quats, axis=0)
                det_bbox[i, :] = np.sum(bboxes, axis=0)

        rot_pos = rot * sc.spatial.as_vectors(
            pos_d["x"].data, pos_d["y"].data, pos_d["z"].data
        )

        _to_spherical(rot_pos, pos_d)

        averaged = sc.groupby(
            pos_d,
            spectrum_dim,
            bins=sc.Variable(
                dims=[spectrum_dim], values=np.arange(-0.5, len(spec_info) + 0.5, 1.0)
            ),
        ).mean("detector")

        sign = averaged["p-sign"].data / sc.abs(averaged["p-sign"].data)
        averaged["p"] = sign * ((np.pi * sc.units.rad) - averaged["p-delta"].data)
        averaged["x"] = (
            averaged["r"].data * sc.sin(averaged["t"].data) * sc.cos(averaged["p"].data)
        )
        averaged["y"] = (
            averaged["r"].data * sc.sin(averaged["t"].data) * sc.sin(averaged["p"].data)
        )
        averaged["z"] = averaged["r"].data * sc.cos(averaged["t"].data)

        pos = sc.spatial.as_vectors(
            averaged["x"].data, averaged["y"].data, averaged["z"].data
        )

        return (
            inv_rot * pos,
            sc.spatial.rotations(dims=[spectrum_dim], values=det_rot_quaternions),
            sc.vectors(dims=[spectrum_dim], values=det_bbox, unit=sc.units.m),
        )
    else:
        pos = np.zeros([nspec, 3])

        for i, spec in enumerate(spec_info):
            if spec.hasDetectors:
                definition = spec_info.getSpectrumDefinition(i)
                n_dets = len(definition)
                vec3s = []
                quats = []
                bboxes = []
                for j in range(n_dets):
                    det_idx = definition[j][0]
                    p = det_info.position(det_idx)
                    r = det_info.rotation(det_idx)
                    vec3s.append([p.X(), p.Y(), p.Z()])
                    quats.append(np.array([r.imagI(), r.imagJ(), r.imagK(), r.real()]))
                    if comp_info.hasValidShape(det_idx):
                        s = comp_info.shape(det_idx)
                        bboxes.append(s.getBoundingBox().width())
                pos[i, :] = np.mean(vec3s, axis=0)
                det_rot_quaternions[i] = np.mean(quats, axis=0)
                det_bbox[i, :] = np.sum(bboxes, axis=0)
            else:
                pos[i, :] = [np.nan, np.nan, np.nan]
                det_rot_quaternions[i] = [np.nan, np.nan, np.nan, np.nan]
                det_bbox[i, :] = [np.nan, np.nan, np.nan]
        return (
            sc.vectors(dims=[spectrum_dim], values=pos, unit=sc.units.m),
            sc.spatial.rotations(dims=[spectrum_dim], values=det_rot_quaternions),
            sc.vectors(
                dims=[spectrum_dim],
                values=det_bbox,
                unit=sc.units.m,
            ),
        )


def _get_dtype_from_values(values, coerce_floats_to_ints):
    if coerce_floats_to_ints and np.all(np.mod(values, 1.0) == 0.0):
        dtype = sc.DType.int32
    elif hasattr(values, 'dtype'):
        dtype = values.dtype
    else:
        if len(values) > 0:
            dtype = type(values[0])
            if dtype is str:
                dtype = sc.DType.string
            elif dtype is int:
                dtype = sc.DType.int64
            elif dtype is float:
                dtype = sc.DType.float64
            else:
                raise RuntimeError(
                    "Cannot handle the dtype that this " "workspace has on Axis 1."
                )
        else:
            raise RuntimeError(
                "Axis 1 of this workspace has no values. " "Cannot determine dtype."
            )
    return dtype


def init_spec_axis(ws):
    axis = ws.getAxis(1)
    dim, unit = validate_and_get_unit(axis.getUnit())
    values = axis.extractValues()
    dtype = _get_dtype_from_values(values, dim == 'spectrum')
    return dim, sc.array(dims=[dim], values=values, unit=unit, dtype=dtype)


def set_bin_masks(bin_masks, dim, index, masked_bins):
    for masked_bin in masked_bins:
        bin_masks['spectrum', index][dim, masked_bin].value = True


def _as_dict_of_variables(d: dict[str, Any]) -> dict[str, sc.Variable]:
    return {
        key: val if isinstance(val, sc.Variable) else sc.scalar(val)
        for key, val in d.items()
    }


def _convert_MatrixWorkspace_info(ws, advanced_geometry=False, load_run_logs=True):
    from mantid.kernel import DeltaEModeType

    common_bins = ws.isCommonBins()
    dim, unit = validate_and_get_unit(ws.getAxis(0).getUnit())
    source_pos, sample_pos = make_component_info(ws)
    spec_dim, spec_coord = init_spec_axis(ws)
    pos, rot, shp = get_detector_properties(
        ws, source_pos, sample_pos, spec_dim, advanced_geometry=advanced_geometry
    )

    coords = {spec_dim: spec_coord}
    # possible x - coord
    if not ws.id() == 'MaskWorkspace':
        if common_bins:
            coords[dim] = sc.Variable(dims=[dim], values=ws.readX(0), unit=unit)
        else:
            coords[dim] = sc.Variable(
                dims=[spec_dim, dim], values=ws.extractX(), unit=unit
            )

    info = {
        "coords": coords,
        "masks": {},
        "attrs": {
            "sample": make_mantid_sample(ws),
            "instrument_name": ws.componentInfo().name(ws.componentInfo().root()),
        },
    }

    if load_run_logs:
        for run_log_name, run_log in process_run_logs(ws):
            info["attrs"][run_log_name] = run_log

    if advanced_geometry:
        info["coords"]["detector_info"] = make_detector_info(ws, spec_dim)

    if not np.all(np.isnan(pos.values)):
        info["coords"].update({"position": pos})

    if rot is not None and shp is not None and not np.all(np.isnan(pos.values)):
        info["attrs"].update({"rotation": rot, "shape": shp})

    if source_pos is not None:
        info["coords"]["source_position"] = source_pos

    if sample_pos is not None:
        info["coords"]["sample_position"] = sample_pos

    if ws.detectorInfo().hasMaskedDetectors():
        spectrum_info = ws.spectrumInfo()
        mask = np.array(
            [spectrum_info.isMasked(i) for i in range(ws.getNumberHistograms())]
        )
        info["masks"][spec_dim] = sc.Variable(dims=[spec_dim], values=mask)

    if ws.getEMode() == DeltaEModeType.Direct:
        info["coords"]["incident_energy"] = _extract_einitial(ws)
    elif ws.getEMode() == DeltaEModeType.Indirect:
        info["coords"]["final_energy"] = _extract_efinal(ws, spec_dim)

    if ws.sample().hasOrientedLattice():
        info["attrs"].update(
            {"sample_ub": make_sample_ub(ws), "sample_u": make_sample_u(ws)}
        )
    return info


def convert_monitors_ws(ws, converter, **ignored):
    spec_dim, spec_coord = init_spec_axis(ws)
    spec_info = ws.spectrumInfo()
    comp_info = ws.componentInfo()
    monitors = []
    spec_indices = (
        (ws.getIndexFromSpectrumNumber(int(i)), i) for i in spec_coord.values
    )
    for index, number in spec_indices:
        definition = spec_info.getSpectrumDefinition(index)
        if not definition.size() == 1:
            raise RuntimeError("Cannot deal with grouped monitor detectors")
        det_index = definition[0][0]  # Ignore time index
        # We only ExtractSpectra for compatibility with
        # existing convert_Workspace2D_to_dataarray. This could instead be
        # refactored if found to be slow
        with run_mantid_alg(
            'ExtractSpectra', InputWorkspace=ws, WorkspaceIndexList=[index]
        ) as monitor_ws:
            # Run logs are already loaded in the data workspace
            single_monitor = converter(monitor_ws, load_run_logs=False)
        # Storing sample_position as an aligned coord of monitors means that monitor
        # data cannot be combined with scattered data even after conversion
        # to wavelength, d-spacing, etc. because conversions of monitors do
        # not use the sample position.
        single_monitor['data'].coords.set_aligned('sample_position', False)
        # Remove redundant information that is duplicated from workspace
        # We get this extra information from the generic converter reuse
        single_monitor['data'].coords.pop('detector_info', None)
        del single_monitor['sample']
        name = comp_info.name(det_index)
        if not comp_info.uniqueName(name):
            name = f'{name}_{number}'
        monitors.append((name, single_monitor))
    return monitors


def convert_monitors_ws_arrays(ws, converter, **ignored):
    spec_dim, spec_coord = init_spec_axis(ws)
    spec_info = ws.spectrumInfo()
    comp_info = ws.componentInfo()
    monitors = []
    spec_indices = (
        (ws.getIndexFromSpectrumNumber(int(i)), i) for i in spec_coord.values
    )
    for index, number in spec_indices:
        definition = spec_info.getSpectrumDefinition(index)
        if not definition.size() == 1:
            raise RuntimeError("Cannot deal with grouped monitor detectors")
        det_index = definition[0][0]  # Ignore time index
        # We only ExtractSpectra for compatibility with
        # existing convert_Workspace2D_to_dataarray. This could instead be
        # refactored if found to be slow
        with run_mantid_alg(
            'ExtractSpectra', InputWorkspace=ws, WorkspaceIndexList=[index]
        ) as monitor_ws:
            # Run logs are already loaded in the data workspace
            single_monitor = converter(monitor_ws, load_run_logs=False)
        # Storing sample_position as a coord of monitors means that monitor
        # data cannot be combined with scattered data even after conversion
        # to wavelength, d-spacing, etc. because conversions of monitors do
        # not use the sample position.
        # But keep it as an attr in case a user needs it.
        get_attrs(single_monitor)['sample_position'] = single_monitor.coords.pop(
            'sample_position'
        )
        # Remove redundant information that is duplicated from workspace
        # We get this extra information from the generic converter reuse
        if 'detector_info' in single_monitor.coords:
            del single_monitor.coords['detector_info']
        del get_attrs(single_monitor)['sample']
        name = comp_info.name(det_index)
        if not comp_info.uniqueName(name):
            name = f'{name}_{number}'
        monitors.append((name, single_monitor))
    return monitors


def convert_Workspace2D_to_data_array(
    ws, load_run_logs=True, advanced_geometry=False, **ignored
):
    warnings.warn(
        'convert_Workspace2D_to_data_array is deprecated in favor of '
        'convert_Workspace2D_to_data_group.',
        VisibleDeprecationWarning,
        stacklevel=4,
    )

    dim, unit = validate_and_get_unit(ws.getAxis(0).getUnit())
    spec_dim, spec_coord = init_spec_axis(ws)

    coords_labs_data = _convert_MatrixWorkspace_info(
        ws, advanced_geometry=advanced_geometry, load_run_logs=load_run_logs
    )
    coords_labs_data["coords"] = _as_dict_of_variables(coords_labs_data["coords"])
    coords_labs_data["attrs"] = _as_dict_of_variables(coords_labs_data["attrs"])
    _, data_unit = validate_and_get_unit(ws.YUnit(), allow_empty=True)
    if ws.id() == 'MaskWorkspace':
        coords_labs_data["data"] = sc.Variable(
            dims=[spec_dim],
            unit=None,
            values=ws.extractY().flatten(),
            dtype=sc.DType.bool,
        )
    else:
        stddev2 = ws.extractE()
        np.multiply(stddev2, stddev2, out=stddev2)  # much faster than np.power
        coords_labs_data["data"] = sc.Variable(
            dims=[spec_dim, dim],
            unit=data_unit,
            values=ws.extractY(),
            variances=stddev2,
        )
    array = sc.DataArray(**coords_labs_data)

    if ws.hasAnyMaskedBins():
        bin_mask = sc.zeros(dims=array.dims, shape=array.shape, dtype=sc.DType.bool)
        for i in range(ws.getNumberHistograms()):
            # maskedBinsIndices throws instead of returning empty list
            if ws.hasMaskedBins(i):
                set_bin_masks(bin_mask, dim, i, ws.maskedBinsIndices(i))
        common_mask = sc.all(bin_mask, spec_dim)
        if sc.identical(common_mask, sc.any(bin_mask, spec_dim)):
            array.masks["bin"] = common_mask
        else:
            array.masks["bin"] = bin_mask

    # Avoid creating dimensions that are not required since this mostly an
    # artifact of inflexible data structures and gets in the way when working
    # with scipp.
    if len(spec_coord.values) == 1:
        if 'position' in array.coords:
            array.coords['position'] = array.coords['position'][spec_dim, 0]
        array = array[spec_dim, 0].copy()
    return array


def convert_Workspace2D_to_data_group(
    ws, load_run_logs=True, advanced_geometry=False, **ignored
) -> sc.DataGroup:
    dim, unit = validate_and_get_unit(ws.getAxis(0).getUnit())
    spec_dim, spec_coord = init_spec_axis(ws)

    coords_labs_data = _convert_MatrixWorkspace_info(
        ws, advanced_geometry=advanced_geometry, load_run_logs=load_run_logs
    )
    _, data_unit = validate_and_get_unit(ws.YUnit(), allow_empty=True)
    if ws.id() == 'MaskWorkspace':
        data = sc.array(
            dims=[spec_dim],
            unit=None,
            values=ws.extractY().flatten(),
            dtype=sc.DType.bool,
        )
    else:
        stddev2 = ws.extractE()
        np.multiply(stddev2, stddev2, out=stddev2)  # much faster than np.power
        data = sc.array(
            dims=[spec_dim, dim],
            unit=data_unit,
            values=ws.extractY(),
            variances=stddev2,
        )
    res = sc.DataGroup(
        {
            "data": sc.DataArray(
                data,
                coords=_as_dict_of_variables(coords_labs_data["coords"]),
                masks=coords_labs_data["masks"],
            ),
            **coords_labs_data["attrs"],
        }
    )

    if ws.hasAnyMaskedBins():
        bin_mask = sc.zeros(dims=data.dims, shape=data.shape, dtype=sc.DType.bool)
        for i in range(ws.getNumberHistograms()):
            # maskedBinsIndices throws instead of returning empty list
            if ws.hasMaskedBins(i):
                set_bin_masks(bin_mask, dim, i, ws.maskedBinsIndices(i))
        common_mask = sc.all(bin_mask, spec_dim)
        if sc.identical(common_mask, sc.any(bin_mask, spec_dim)):
            res["data"].masks["bin"] = common_mask
        else:
            res["data"].masks["bin"] = bin_mask

    # Avoid creating dimensions that are not required since this mostly an
    # artifact of inflexible data structures and gets in the way when working
    # with scipp.
    if len(spec_coord.values) == 1:
        if 'position' in res["data"].coords:
            res["data"].coords['position'] = res["data"].coords['position'][spec_dim, 0]
        res["data"] = res["data"][spec_dim, 0].copy()
    return res


def _contains_weighted_events(spectrum) -> bool:
    from mantid.api import EventType

    return spectrum.getEventType() in (EventType.WEIGHTED, EventType.WEIGHTED_NOTIME)


def convert_EventWorkspace_to_data_array(
    ws, load_pulse_times=True, advanced_geometry=False, load_run_logs=True, **ignored
):
    warnings.warn(
        'convert_EventWorkspace_to_data_array is deprecated in favor of '
        'convert_EventWorkspace_to_data_group.',
        VisibleDeprecationWarning,
        stacklevel=4,
    )

    dim, unit = validate_and_get_unit(ws.getAxis(0).getUnit())
    spec_dim, spec_coord = init_spec_axis(ws)
    nHist = ws.getNumberHistograms()
    _, data_unit = validate_and_get_unit(ws.YUnit(), allow_empty=True)

    n_event = ws.getNumberEvents()
    coord = sc.empty(dims=['event'], shape=[n_event], unit=unit, dtype=sc.DType.float64)
    weights = sc.ones(
        dims=['event'],
        shape=[n_event],
        unit=data_unit,
        dtype=sc.DType.float32,
        with_variances=True,
    )
    pulse_times = (
        sc.empty(
            dims=['event'], shape=[n_event], dtype=sc.DType.datetime64, unit=sc.units.ns
        )
        if load_pulse_times
        else None
    )

    begins = sc.zeros(
        dims=[spec_dim, dim], shape=[nHist, 1], dtype=sc.DType.int64, unit=None
    )
    ends = begins.copy()
    if n_event > 0:  # Skip expensive loop if there are no events
        current = 0
        for i in range(nHist):
            sp = ws.getSpectrum(i)
            size = sp.getNumberEvents()
            begins.values[i] = current
            ends.values[i] = current + size
            if size == 0:  # Skip expensive getters
                continue
            coord['event', current : current + size].values = sp.getTofs()
            if load_pulse_times:
                pulse_times[
                    'event', current : current + size
                ].values = sp.getPulseTimesAsNumpy()
            if _contains_weighted_events(sp):
                weights['event', current : current + size].values = sp.getWeights()
                weights[
                    'event', current : current + size
                ].variances = sp.getWeightErrors()
            current += size

    proto_events = {'data': weights, 'coords': {dim: coord}}
    if load_pulse_times:
        proto_events["coords"]["pulse_time"] = pulse_times
    events = sc.DataArray(**proto_events)

    coords_labs_data = _convert_MatrixWorkspace_info(
        ws, advanced_geometry=advanced_geometry, load_run_logs=load_run_logs
    )
    coords_labs_data["coords"] = _as_dict_of_variables(coords_labs_data["coords"])
    coords_labs_data["attrs"] = _as_dict_of_variables(coords_labs_data["attrs"])
    # For now we ignore potential finer bin edges to avoid creating too many
    # bins. Use just a single bin along dim and use extents given by workspace
    # edges.
    # TODO If there are events outside edges this might create bins with
    # events that are not within bin bounds. Consider using `bin` instead
    # of `bins`?
    edges = coords_labs_data['coords'][dim]
    # Using range slice of thickness 1 to avoid transposing 2-D coords
    coords_labs_data['coords'][dim] = sc.concat([edges[dim, :1], edges[dim, -1:]], dim)
    coords_labs_data["data"] = sc.bins(begin=begins, end=ends, dim='event', data=events)
    return sc.DataArray(**coords_labs_data)


def convert_EventWorkspace_to_data_group(
    ws, load_pulse_times=True, advanced_geometry=False, load_run_logs=True, **ignored
):
    dim, unit = validate_and_get_unit(ws.getAxis(0).getUnit())
    spec_dim, spec_coord = init_spec_axis(ws)
    nHist = ws.getNumberHistograms()
    _, data_unit = validate_and_get_unit(ws.YUnit(), allow_empty=True)

    n_event = ws.getNumberEvents()
    coord = sc.empty(dims=['event'], shape=[n_event], unit=unit, dtype=sc.DType.float64)
    weights = sc.ones(
        dims=['event'],
        shape=[n_event],
        unit=data_unit,
        dtype=sc.DType.float32,
        with_variances=True,
    )
    pulse_times = (
        sc.empty(
            dims=['event'], shape=[n_event], dtype=sc.DType.datetime64, unit=sc.units.ns
        )
        if load_pulse_times
        else None
    )

    begins = sc.zeros(
        dims=[spec_dim, dim], shape=[nHist, 1], dtype=sc.DType.int64, unit=None
    )
    ends = begins.copy()
    if n_event > 0:  # Skip expensive loop if there are no events
        current = 0
        for i in range(nHist):
            sp = ws.getSpectrum(i)
            size = sp.getNumberEvents()
            begins.values[i] = current
            ends.values[i] = current + size
            if size == 0:  # Skip expensive getters
                continue
            coord['event', current : current + size].values = sp.getTofs()
            if load_pulse_times:
                pulse_times[
                    'event', current : current + size
                ].values = sp.getPulseTimesAsNumpy()
            if _contains_weighted_events(sp):
                weights['event', current : current + size].values = sp.getWeights()
                weights[
                    'event', current : current + size
                ].variances = sp.getWeightErrors()
            current += size

    proto_events = {'data': weights, 'coords': {dim: coord}}
    if load_pulse_times:
        proto_events["coords"]["pulse_time"] = pulse_times
    events = sc.DataArray(**proto_events)

    coords_labs_data = _convert_MatrixWorkspace_info(
        ws, advanced_geometry=advanced_geometry, load_run_logs=load_run_logs
    )
    # For now we ignore potential finer bin edges to avoid creating too many
    # bins. Use just a single bin along dim and use extents given by workspace
    # edges.
    # TODO If there are events outside edges this might create bins with
    # events that are not within bin bounds. Consider using `bin` instead
    # of `bins`?
    edges = coords_labs_data['coords'][dim]
    # Using range slice of thickness 1 to avoid transposing 2-D coords
    coords_labs_data['coords'][dim] = sc.concat([edges[dim, :1], edges[dim, -1:]], dim)

    data = sc.bins(begin=begins, end=ends, dim='event', data=events)
    return sc.DataGroup(
        {
            "data": sc.DataArray(
                data, coords=_as_dict_of_variables(coords_labs_data["coords"])
            ),
            **coords_labs_data["attrs"],
        }
    )


def convert_MDHistoWorkspace_to_data_group(md_histo, **ignored) -> sc.DataGroup:
    ndims = md_histo.getNumDims()
    coords = {}
    dims_used = []
    for i in range(ndims):
        dim = md_histo.getDimension(i)
        frame = dim.getMDFrame()
        sc_dim = md_dimension(dim, i)
        coords[sc_dim] = sc.array(
            dims=[sc_dim],
            values=np.linspace(dim.getMinimum(), dim.getMaximum(), dim.getNBins()),
            unit=md_unit(frame),
        )
        dims_used.append(sc_dim)
    data = sc.array(
        dims=dims_used,
        values=md_histo.getSignalArray(),
        variances=md_histo.getErrorSquaredArray(),
        unit=sc.units.counts,
    )
    nevents = sc.array(dims=dims_used, values=md_histo.getNumEventsArray())
    return sc.DataGroup(
        {'data': sc.DataArray(coords=coords, data=data), 'nevents': nevents}
    )


def convert_MDHistoWorkspace_to_data_array(md_histo, **ignored):
    warnings.warn(
        'convert_MDHistoWorkspace_to_data_array is deprecated in favor of '
        'convert_MDHistoWorkspace_to_data_group.',
        VisibleDeprecationWarning,
        stacklevel=4,
    )

    ndims = md_histo.getNumDims()
    coords = {}
    dims_used = []
    for i in range(ndims):
        dim = md_histo.getDimension(i)
        frame = dim.getMDFrame()
        sc_dim = md_dimension(dim, i)
        coords[sc_dim] = sc.Variable(
            dims=[sc_dim],
            values=np.linspace(dim.getMinimum(), dim.getMaximum(), dim.getNBins()),
            unit=md_unit(frame),
        )
        dims_used.append(sc_dim)
    data = sc.Variable(
        dims=dims_used,
        values=md_histo.getSignalArray(),
        variances=md_histo.getErrorSquaredArray(),
        unit=sc.units.counts,
    )
    nevents = sc.Variable(dims=dims_used, values=md_histo.getNumEventsArray())
    return sc.DataArray(coords=coords, data=data, attrs={'nevents': nevents})


def convert_TableWorkspace_to_dataset(ws, error_connection=None, **ignored):
    """
    Converts from a Mantid TableWorkspace to a scipp dataset.

    It is possible to assign a column as the error for another column,
    in which case data from the two columns will be represented by a single scipp
    variable with variance. This is done using the error_connection Keyword
    argument. The error is transformed to variance in this converter.

    Parameters
    ----------
    ws:
        Mantid TableWorkspace to be converted into scipp dataset.
    error_connection:
        Dict with data column names as keys to names of their error column.
    """

    # Extract information from workspace
    n_columns = ws.columnCount()
    columnNames = ws.getColumnNames()  # list of names matching each column
    columnTypes = ws.columnTypes()  # list of types matching each column

    # Types available in TableWorkspace that can not be loaded into scipp
    blacklist_types = []
    # Types for which the transformation from error to variance will fail
    blacklist_variance_types = ["str"]

    result = {}
    for i in range(n_columns):
        if columnTypes[i] in blacklist_types:
            continue  # skips loading data of this type

        data_name = columnNames[i]
        if error_connection is None:
            result[data_name] = sc.Variable(dims=['row'], values=ws.column(i))
        elif data_name in error_connection:
            # This data has error available
            error_name = error_connection[data_name]
            error_index = columnNames.index(error_name)

            if columnTypes[error_index] in blacklist_variance_types:
                # Raise error to avoid numpy square error for strings
                raise RuntimeError(
                    "Variance can not have type string. \n"
                    + "Data:     "
                    + str(data_name)
                    + "\n"
                    + "Variance: "
                    + str(error_name)
                    + "\n"
                )

            variance = np.array(ws.column(error_name)) ** 2
            result[data_name] = sc.Variable(
                dims=['row'], values=np.array(ws.column(i)), variances=variance
            )
        elif data_name not in error_connection.values():
            # This data is not an error for another dataset, and has no error
            result[data_name] = sc.Variable(dims=['row'], values=ws.column(i))

    return sc.Dataset(result) if result else sc.Dataset({})


def convert_WorkspaceGroup_to_data_group(group_workspace, **kwargs):
    workspace_dict = sc.DataGroup()
    for i in range(group_workspace.getNumberOfEntries()):
        workspace = group_workspace.getItem(i)
        workspace_name = (
            workspace.name().replace(f'{group_workspace.name()}', '').strip('_')
        )
        workspace_dict[workspace_name] = from_mantid(workspace, **kwargs)

    return workspace_dict


def convert_WorkspaceGroup_to_dataarray_dict(group_workspace, **kwargs):
    warnings.warn(
        'convert_WorkspaceGroup_to_dataarray_dict is deprecated in favor of '
        'convert_WorkspaceGroup_to_data_group.',
        VisibleDeprecationWarning,
        stacklevel=4,
    )

    workspace_dict = {}
    for i in range(group_workspace.getNumberOfEntries()):
        workspace = group_workspace.getItem(i)
        workspace_name = (
            workspace.name().replace(f'{group_workspace.name()}', '').strip('_')
        )
        workspace_dict[workspace_name] = from_mantid(workspace, **kwargs)

    return workspace_dict


def from_mantid(workspace, **kwargs) -> sc.DataGroup:
    """Convert Mantid workspace to a scipp data group.

    Parameters
    ----------
    workspace:
        Mantid workspace to convert.
    kwargs:
        Forwarded to the workspace-specific converter.

    Returns
    -------
        A data group representing ``workspace``.
    """
    monitor_ws = None
    workspaces_to_delete = []
    w_id = workspace.id()
    if w_id == 'Workspace2D' or w_id == 'RebinnedOutput' or w_id == 'MaskWorkspace':
        n_monitor = 0
        spec_info = workspace.spectrumInfo()
        for i in range(len(spec_info)):
            if spec_info.hasDetectors(i) and spec_info.isMonitor(i):
                n_monitor += 1
        # If there are *only* monitors we do not move them to an attribute
        if n_monitor > 0 and n_monitor < len(spec_info):
            import mantid.simpleapi as mantid

            workspace, monitor_ws = mantid.ExtractMonitors(workspace)
            workspaces_to_delete.append(workspace)
            workspaces_to_delete.append(monitor_ws)
        scipp_obj = convert_Workspace2D_to_data_group(workspace, **kwargs)
    elif w_id == 'EventWorkspace':
        scipp_obj = convert_EventWorkspace_to_data_group(workspace, **kwargs)
    elif w_id == 'TableWorkspace':
        scipp_obj = convert_TableWorkspace_to_dataset(workspace, **kwargs)
    elif w_id == 'MDHistoWorkspace':
        scipp_obj = convert_MDHistoWorkspace_to_data_group(workspace, **kwargs)
    elif w_id == 'WorkspaceGroup':
        scipp_obj = convert_WorkspaceGroup_to_data_group(workspace, **kwargs)
    else:
        raise RuntimeError(f'Unsupported workspace type {w_id}')

    # TODO Is there ever a case where a Workspace2D has a separate monitor
    # workspace? This is not handled by ExtractMonitors above, I think.
    if monitor_ws is None:
        if hasattr(workspace, 'getMonitorWorkspace'):
            try:
                monitor_ws = workspace.getMonitorWorkspace()
            except RuntimeError:
                # Have to try/fail here. No inspect method on Mantid for this.
                pass

    if monitor_ws is not None:
        if monitor_ws.id() == 'MaskWorkspace' or monitor_ws.id() == 'Workspace2D':
            converter = convert_Workspace2D_to_data_group
        elif monitor_ws.id() == 'EventWorkspace':
            converter = convert_EventWorkspace_to_data_group
        scipp_obj["monitors"] = sc.DataGroup(
            convert_monitors_ws(monitor_ws, converter, **kwargs)
        )
    for ws in workspaces_to_delete:
        mantid.DeleteWorkspace(ws)

    return scipp_obj


def array_from_mantid(workspace, **kwargs) -> sc.DataArray | sc.Dataset:
    """Convert Mantid workspace to a scipp data array or dataset.
    :param workspace: Mantid workspace to convert.

    .. deprecated:: 23.07.0
       The old ``from_mantid`` was renamed to ``array_from_mantid`` and is deprecated
       in favor of the new ``from_mantid`` which return a data group.
    """
    warnings.warn(
        "scippneutron.array_from_mantid has been deprecated and is scheduled for "
        "removal in scippneutron v24.01.0. "
        "Use the new scippneutron.from_mantid instead.",
        VisibleDeprecationWarning,
        stacklevel=4,
    )

    scipp_obj = None  # This is either a Dataset or DataArray
    monitor_ws = None
    workspaces_to_delete = []
    w_id = workspace.id()
    if w_id == 'Workspace2D' or w_id == 'RebinnedOutput' or w_id == 'MaskWorkspace':
        n_monitor = 0
        spec_info = workspace.spectrumInfo()
        for i in range(len(spec_info)):
            if spec_info.hasDetectors(i) and spec_info.isMonitor(i):
                n_monitor += 1
        # If there are *only* monitors we do not move them to an attribute
        if n_monitor > 0 and n_monitor < len(spec_info):
            import mantid.simpleapi as mantid

            workspace, monitor_ws = mantid.ExtractMonitors(workspace)
            workspaces_to_delete.append(workspace)
            workspaces_to_delete.append(monitor_ws)
        scipp_obj = convert_Workspace2D_to_data_array(workspace, **kwargs)
    elif w_id == 'EventWorkspace':
        scipp_obj = convert_EventWorkspace_to_data_array(workspace, **kwargs)
    elif w_id == 'TableWorkspace':
        scipp_obj = convert_TableWorkspace_to_dataset(workspace, **kwargs)
    elif w_id == 'MDHistoWorkspace':
        scipp_obj = convert_MDHistoWorkspace_to_data_array(workspace, **kwargs)
    elif w_id == 'WorkspaceGroup':
        scipp_obj = convert_WorkspaceGroup_to_dataarray_dict(workspace, **kwargs)

    if scipp_obj is None:
        raise RuntimeError(f'Unsupported workspace type {w_id}')

    # TODO Is there ever a case where a Workspace2D has a separate monitor
    # workspace? This is not handled by ExtractMonitors above, I think.
    if monitor_ws is None:
        if hasattr(workspace, 'getMonitorWorkspace'):
            try:
                monitor_ws = workspace.getMonitorWorkspace()
            except RuntimeError:
                # Have to try/fail here. No inspect method on Mantid for this.
                pass

    if monitor_ws is not None:
        if monitor_ws.id() == 'MaskWorkspace' or monitor_ws.id() == 'Workspace2D':
            converter = convert_Workspace2D_to_data_array
        elif monitor_ws.id() == 'EventWorkspace':
            converter = convert_EventWorkspace_to_data_array

        monitors = convert_monitors_ws_arrays(monitor_ws, converter, **kwargs)
        for name, monitor in monitors:
            get_attrs(scipp_obj)[name] = sc.scalar(monitor)
    for ws in workspaces_to_delete:
        mantid.DeleteWorkspace(ws)

    return scipp_obj


def load_with_mantid(
    filename: str | Path = "",
    load_pulse_times=True,
    instrument_filename=None,
    error_connection=None,
    mantid_alg='Load',
    mantid_args=None,
    advanced_geometry=False,
) -> sc.DataGroup:
    """Load a file using Mantid.

    Wraps Mantid's loaders and converts the result to a scipp data group.

    See also the neutron-data tutorial.

    Note that this function requires mantid to be installed and available in
    the same Python environment as scipp.

    Parameters
    ----------
    filename:
        The name of the Nexus/HDF file to be loaded.
    load_pulse_times:
        Read the pulse times if True.
    instrument_filename:
        If specified, over-write the instrument definition
        in the final dataset with the geometry contained in the file.
    error_connection:
        Dict with data column names as keys to names of their error column.
        Only used when the loaded workspace is a TableWorkspace.
        See scippneutron.mantid.convert_TableWorkspace_to_dataset
    mantid_alg:
        Mantid algorithm to use for loading. Default is `Load`.
    mantid_args:
        Dict of keyword arguments to forward to Mantid.
    advanced_geometry:
        If True, load the full detector geometry including shapes and rotations.
        The positions of grouped detectors are spherically averaged.
        If False, load only the detector position, and return
        the cartesian average of the grouped detector positions.

    Returns
    -------
    :
        A Data group containing the neutron event/histogram data and the
        instrument geometry.

    Raises
    ------
    RuntimeError
        If the Mantid workspace type returned by the Mantid loader is not
        either EventWorkspace or Workspace2D.

    Examples
    --------
    >>> from scippneutron import load_with_mantid
    >>> load_with_mantid(filename='PG3_4844_event.nxs',
    ...                  load_pulse_times=False,
    ...                  mantid_args={
    ...                      'BankName': 'bank184',
    ...                      'LoadMonitors': True})  # doctest: +SKIP
    """

    if mantid_args is None:
        mantid_args = {}

    _check_file_path(str(filename), mantid_alg)

    with run_mantid_alg(mantid_alg, str(filename), **mantid_args) as loaded:
        # Determine what Load has provided us
        from mantid.api import Workspace

        if isinstance(loaded, Workspace):
            # A single workspace
            data_ws = loaded
        else:
            # Separate data and monitor workspaces
            data_ws = loaded.OutputWorkspace

        if instrument_filename is not None:
            import mantid.simpleapi as mantid

            mantid.LoadInstrument(
                data_ws, FileName=instrument_filename, RewriteSpectraMap=True
            )
        return from_mantid(
            data_ws,
            load_pulse_times=load_pulse_times,
            error_connection=error_connection,
            advanced_geometry=advanced_geometry,
        )


def load(
    filename: str | Path = "",
    load_pulse_times=True,
    instrument_filename=None,
    error_connection=None,
    mantid_alg='Load',
    mantid_args=None,
    advanced_geometry=False,
) -> sc.Dataset:
    """Load a file using Mantid.

    Wraps Mantid's loaders and converts the result to a scipp dataset.

    See also the neutron-data tutorial.

    Note that this function requires mantid to be installed and available in
    the same Python environment as scipp.

    Parameters
    ----------
    filename:
        The name of the Nexus/HDF file to be loaded.
    load_pulse_times:
        Read the pulse times if True.
    instrument_filename:
        If specified, over-write the instrument definition
        in the final dataset with the geometry contained in the file.
    error_connection:
        Dict with data column names as keys to names of their error column.
        Only used when the loaded workspace is a TableWorkspace.
        See scippneutron.mantid.convert_TableWorkspace_to_dataset
    mantid_alg:
        Mantid algorithm to use for loading. Default is `Load`.
    mantid_args:
        Dict of keyword arguments to forward to Mantid.
    advanced_geometry:
        If True, load the full detector geometry including shapes and rotations.
        The positions of grouped detectors are spherically averaged.
        If False, load only the detector position, and return
        the cartesian average of the grouped detector positions.

    Returns
    -------
    :
        A Dataset containing the neutron event/histogram data and the
        instrument geometry.

    Raises
    ------
    RuntimeError
        If the Mantid workspace type returned by the Mantid loader is not
        either EventWorkspace or Workspace2D.

    Examples
    --------
    >>> from scippneutron import load_with_mantid
    >>> d = sc.Dataset()
    >>> d["sample"] = load(filename='PG3_4844_event.nxs',
    ...                    load_pulse_times=False,
    ...                    mantid_args={'BankName': 'bank184',
    ...                                 'LoadMonitors': True})  # doctest: +SKIP

    .. deprecated:: 23.07.0
       Use load_with_mantid instead.
    """
    warnings.warn(
        "scippneutron.load has been deprecated and is scheduled for removal in "
        "scippneutron v24.01.0. Use the new scippneutron.load_with_mantid instead.",
        VisibleDeprecationWarning,
        stacklevel=2,
    )

    if mantid_args is None:
        mantid_args = {}

    _check_file_path(str(filename), mantid_alg)

    with run_mantid_alg(mantid_alg, str(filename), **mantid_args) as loaded:
        # Determine what Load has provided us
        from mantid.api import Workspace

        if isinstance(loaded, Workspace):
            # A single workspace
            data_ws = loaded
        else:
            # Separate data and monitor workspaces
            data_ws = loaded.OutputWorkspace

        if instrument_filename is not None:
            import mantid.simpleapi as mantid

            mantid.LoadInstrument(
                data_ws, FileName=instrument_filename, RewriteSpectraMap=True
            )
        return array_from_mantid(
            data_ws,
            load_pulse_times=load_pulse_times,
            error_connection=error_connection,
            advanced_geometry=advanced_geometry,
        )


def _is_mantid_loadable(filename):
    from mantid.api import FileFinder

    if FileFinder.getFullPath(filename):
        return True
    else:
        try:
            # findRuns throws rather than return empty so need try-catch
            FileFinder.findRuns(filename)
            return True
        except Exception:
            return False


def _check_file_path(filename, mantid_alg):
    from mantid.api import AlgorithmManager, FileProperty, FrameworkManager

    FrameworkManager.Instance()
    alg = AlgorithmManager.createUnmanaged(mantid_alg)
    filename_property = [
        prop for prop in alg.getProperties() if isinstance(prop, FileProperty)
    ]
    # Only with FileProperty can Mantid take fuzzy matches to filenames and run numbers
    # If the top level Load algorithm is requested (has no properties of its own)
    # we know that it's child algorithm uses FileProperty. If the child algorithm
    # is called directly we attempt to find FileProperty. Otherwise paths should be
    # absolute
    if filename_property or mantid_alg == 'Load':
        if not _is_mantid_loadable(filename):
            raise ValueError(
                f"Mantid cannot find {filename} and therefore will not load it."
            ) from None
    else:
        if not os.path.isfile(filename):
            raise ValueError(
                f"Cannot find file {filename} and therefore will not load it."
            )


def validate_dim_and_get_mantid_string(unit_dim):
    known_units = {
        'energy_transfer': "DeltaE",
        'tof': "TOF",
        'wavelength': "Wavelength",
        'energy': "Energy",
        'dspacing': "dSpacing",
        'Q': "MomentumTransfer",
        'Q^2': "QSquared",
    }

    user_k = str(unit_dim).casefold()
    known_units = {k.casefold(): v for k, v in known_units.items()}

    if user_k not in known_units:
        raise RuntimeError(
            "Axis unit not currently supported."
            f"Possible values are: {list(known_units.keys())}, "
            f"got '{unit_dim}'. "
        )
    else:
        return known_units[user_k]


def to_mantid(data, dim, instrument_file=None):
    """
    Convert data to a Mantid workspace.

    The Mantid layout expect the spectra to be the Outer-most dimension,
    i.e. y.shape[0]. If that is not the case you might have to transpose
    your data to fit that, otherwise it will not be aligned correctly in the
    Mantid workspace.

    :param data: Data to be converted.
    :param dim: Coord to use for Mantid's first axis (X).
    :param instrument_file: Instrument file that will be
                            loaded into the workspace
    :returns: Workspace containing converted data. The concrete workspace type
              may differ depending on the content of `data`.
    """
    if not isinstance(data, sc.DataArray):
        raise RuntimeError(
            "Currently only data arrays can be converted to a Mantid workspace"
        )
    try:
        import mantid.simpleapi as mantid
    except ImportError:
        raise ImportError(
            "Mantid Python API was not found, please install Mantid framework "
            "as detailed in the installation instructions (https://scipp."
            "github.io/getting-started/installation.html)"
        ) from None
    x = data.coords[dim].values
    y = data.values
    e = data.variances

    if len(y.shape) not in (1, 2):
        raise ValueError("Currently can only handle 2D data.")

    e = np.sqrt(e) if e is not None else np.sqrt(y)

    # Convert a single array (e.g. single spectra) into 2d format
    if len(y.shape) == 1:
        y = np.array([y])

    if len(e.shape) == 1:
        e = np.array([e])

    unitX = validate_dim_and_get_mantid_string(dim)

    nspec = y.shape[0]
    if len(x.shape) == 1:
        # SCIPP is using a  1:n spectra coord mapping, Mantid needs
        # a 1:1 mapping so expand this out
        x = np.broadcast_to(x, shape=(nspec, len(x)))

    nbins = x.shape[1]
    nitems = y.shape[1]

    ws = mantid.WorkspaceFactory.create(
        "Workspace2D", NVectors=nspec, XLength=nbins, YLength=nitems
    )
    if data.unit != sc.units.counts:
        ws.setDistribution(True)

    for i in range(nspec):
        ws.setX(i, x[i])
        ws.setY(i, y[i])
        ws.setE(i, e[i])

    # Set X-Axis unit
    ws.getAxis(0).setUnit(unitX)

    if instrument_file is not None:
        mantid.LoadInstrument(ws, FileName=instrument_file, RewriteSpectraMap=True)

    return ws


def _table_to_data_array(table, key, value, stddev):
    stddevs = table[stddev].values
    dim = 'parameter'
    coord = table[key].data.copy().rename_dims({'row': dim})
    return sc.DataArray(
        data=sc.Variable(
            dims=[dim], values=table[value].values, variances=stddevs * stddevs
        ),
        coords={dim: coord},
    )


def _fit_workspace(ws, mantid_args):
    """
    Performs a fit on the workspace.

    :param ws: The workspace on which the fit will be performed
    :returns: Dataset containing all of Fit's outputs
    """
    with run_mantid_alg(
        'Fit', InputWorkspace=ws, **mantid_args, CreateOutput=True
    ) as fit:
        # This is assuming that all parameters are dimensionless. If this is
        # not the case we should use a dataset with a scalar variable per
        # parameter instead. Or better, a dict of scalar variables?
        parameters = convert_TableWorkspace_to_dataset(fit.OutputParameters)
        parameters = _table_to_data_array(
            parameters, key='Name', value='Value', stddev='Error'
        )
        out = convert_Workspace2D_to_data_array(fit.OutputWorkspace).drop_coords(
            'empty'
        )
        data = {}
        data['data'] = out['empty', 0]
        data['calculated'] = out['empty', 1]
        data['diff'] = out['empty', 2]
        parameters.coords['status'] = sc.scalar(fit.OutputStatus)
        parameters.coords['chi^2/d.o.f.'] = sc.scalar(fit.OutputChi2overDoF)
        parameters.coords['function'] = sc.scalar(str(fit.Function))
        parameters.coords['cost_function'] = sc.scalar(fit.CostFunction)
        return parameters, sc.Dataset(data)


def fit(data, mantid_args):
    if len(data.dims) != 1 or 'WorkspaceIndex' in mantid_args:
        raise RuntimeError(
            "Only 1D fitting is supported. Use scipp slicing and do not"
            "provide a WorkspaceIndex."
        )
    dim = data.dims[0]
    ws = to_mantid(data, dim)
    mantid_args['workspace_index'] = 0
    return _fit_workspace(ws, mantid_args)


def _try_except(op, possible_except, failure, **kwargs):
    try:
        return op(**kwargs)
    except possible_except:
        return failure


def _get_instrument_efixed(workspace):
    inst = workspace.getInstrument()
    if inst.hasParameter('Efixed'):
        return inst.getNumberParameter('EFixed')[0]

    if inst.hasParameter('analyser'):
        analyser_name = inst.getStringParameter('analyser')[0]
        analyser_comp = inst.getComponentByName(analyser_name)

        if analyser_comp is not None and analyser_comp.hasParameter('Efixed'):
            return analyser_comp.getNumberParameter('EFixed')[0]

    return None


def _extract_einitial(ws):
    if ws.run().hasProperty("Ei"):
        ei = ws.run().getProperty("Ei").value
    elif ws.run().hasProperty('EnergyRequest'):
        ei = ws.run().getProperty('EnergyRequest').value[-1]
    else:
        ei = 0
    return sc.scalar(ei, unit=sc.Unit("meV"))


def _extract_efinal(ws, spec_dim):
    detInfo = ws.detectorInfo()
    specInfo = ws.spectrumInfo()
    ef = np.empty(shape=(specInfo.size(),), dtype=float)
    ef[:] = np.nan
    analyser_ef = _get_instrument_efixed(workspace=ws)
    ids = detInfo.detectorIDs()
    for spec_index in range(len(specInfo)):
        detector_ef = None
        if specInfo.hasDetectors(spec_index):
            # Just like mantid, we only take the first entry of the group.
            det_index = specInfo.getSpectrumDefinition(spec_index)[0][0]
            detector_ef = _try_except(
                op=ws.getEFixed,
                possible_except=RuntimeError,
                failure=None,
                detId=int(ids[det_index]),
            )
        detector_ef = detector_ef if detector_ef is not None else analyser_ef
        if not detector_ef:
            continue  # Cannot assign an Ef. May or may not be an error
            # - i.e. a diffraction detector, monitor etc.
        ef[spec_index] = detector_ef

    return sc.Variable(dims=[spec_dim], values=ef, unit=sc.Unit("meV"))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

"""Chopper utilities."""

from .disk_chopper import DiskChopper, DiskChopperType
from .filtering import collapse_plateaus, filter_in_phase, find_plateaus
from .nexus_chopper import extract_chopper_from_nexus

__all__ = [
    'DiskChopper',
    'DiskChopperType',
    'collapse_plateaus',
    'filter_in_phase',
    'find_plateaus',
    'extract_chopper_from_nexus',
]



from mantid.simpleapi import Load, LoadDiffCal

import scippneutron as scn

ws = Load('PG3_4844_event.nxs')
ws = LoadDiffCal('PG3_golden.cal', InputWorkspace='ws', WorkspaceName='ws')
cal = scn.from_mantid(ws[0]).rename_dims({'row': 'spectrum'})
cal['tzero'].unit = 'us'
cal['difc'].unit = 'us/angstrom'
cal['difa'].unit = 'us/(angstrom*angstrom)'
cal.to_hdf5('PG3_4844_calibration.h5')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

"""Tests of package integrity.

Note that additional imports need to be added for repositories that
contain multiple packages.
"""

import scippneutron as pkg


def test_has_version():
    assert hasattr(pkg, '__version__')


# This is for CI package tests. They need to run tests with minimal dependencies,
# that is, without installing pytest. This code does not affect pytest.
if __name__ == '__main__':
    test_has_version()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import numpy as np
import scipp as sc
import scipp.testing

from scippneutron.peaks import FitAssessment, FitResult, remove_peaks
from scippneutron.peaks.model import GaussianModel, PolynomialModel


def test_remove_peaks_single_peak_clean():
    background = PolynomialModel(degree=1, prefix='bkg_')
    peak = GaussianModel(prefix='peak_')

    a0 = sc.scalar(140.0, unit='counts')
    a1 = sc.scalar(-6.1, unit='counts/Å')
    amplitude = sc.scalar(0.6, unit='counts*Å')
    loc = sc.scalar(1.5, unit='Å')
    scale = sc.scalar(0.04, unit='Å')
    params = {
        'bkg_a0': a0,
        'bkg_a1': a1,
        'peak_amplitude': amplitude,
        'peak_loc': loc,
        'peak_scale': scale,
    }

    x = sc.linspace('d', 0.2, 2.4, 200, unit='Å')
    y_bkg = background(x, bkg_a0=a0, bkg_a1=a1)
    y = y_bkg + peak(x, peak_amplitude=amplitude, peak_loc=loc, peak_scale=scale)
    with_peak = sc.DataArray(y, coords={'d': x})
    original = with_peak.copy()
    expected = sc.DataArray(y_bkg, coords={'d': x})

    result = FitResult(
        aic=sc.scalar(np.nan),
        assessment=FitAssessment.success,
        background=background,
        message='',
        p_value=sc.scalar(np.nan),
        peak=peak,
        popt=params,
        red_chisq=sc.scalar(np.nan),
        window=sc.array(dims=['range'], values=[1.2, 1.7], unit='Å'),
    )
    without_peak = remove_peaks(with_peak, [result])

    sc.testing.assert_identical(with_peak, original)
    sc.testing.assert_allclose(without_peak, expected, rtol=sc.scalar(1e-6))


def test_remove_peaks_single_peak_noisy():
    rng = np.random.default_rng(8381)

    background = PolynomialModel(degree=1, prefix='bkg_')
    peak = GaussianModel(prefix='peak_')

    a0 = sc.scalar(140.0, unit='counts')
    a1 = sc.scalar(-6.1, unit='counts/Å')
    amplitude = sc.scalar(0.6, unit='counts*Å')
    loc = sc.scalar(1.5, unit='Å')
    scale = sc.scalar(0.04, unit='Å')
    params = {
        'bkg_a0': a0,
        'bkg_a1': a1,
        'peak_amplitude': amplitude,
        'peak_loc': loc,
        'peak_scale': scale,
    }

    x = sc.linspace('d', 0.2, 2.4, 200, unit='Å')
    y_bkg = background(x, bkg_a0=a0, bkg_a1=a1) + sc.array(
        dims=x.dims, values=rng.normal(0.0, 0.3, x.shape), unit='counts'
    )
    y = y_bkg + peak(x, peak_amplitude=amplitude, peak_loc=loc, peak_scale=scale)
    with_peak = sc.DataArray(y, coords={'d': x})
    original = with_peak.copy()
    expected = sc.DataArray(y_bkg, coords={'d': x})

    result = FitResult(
        aic=sc.scalar(np.nan),
        assessment=FitAssessment.success,
        background=background,
        message='',
        p_value=sc.scalar(np.nan),
        peak=peak,
        popt=params,
        red_chisq=sc.scalar(np.nan),
        window=sc.array(dims=['range'], values=[1.2, 1.7], unit='Å'),
    )
    without_peak = remove_peaks(with_peak, [result])

    sc.testing.assert_identical(with_peak, original)
    sc.testing.assert_allclose(without_peak, expected, rtol=sc.scalar(1e-6))


def test_remove_peaks_two_peak_noisy():
    rng = np.random.default_rng(8381)

    background = PolynomialModel(degree=1, prefix='bkg_')
    peak = GaussianModel(prefix='peak_')
    a0 = sc.scalar(110.0, unit='counts')
    a1 = sc.scalar(-5.7, unit='counts/Å')
    amplitude1 = sc.scalar(0.6, unit='counts*Å')
    loc1 = sc.scalar(1.8, unit='Å')
    scale1 = sc.scalar(0.04, unit='Å')
    amplitude2 = sc.scalar(0.3, unit='counts*Å')
    loc2 = sc.scalar(0.9, unit='Å')
    scale2 = sc.scalar(0.05, unit='Å')
    params1 = {
        'bkg_a0': a0,
        'bkg_a1': a1,
        'peak_amplitude': amplitude1,
        'peak_loc': loc1,
        'peak_scale': scale1,
    }
    params2 = {
        'bkg_a0': a0,
        'bkg_a1': a1,
        'peak_amplitude': amplitude2,
        'peak_loc': loc2,
        'peak_scale': scale2,
    }

    x = sc.linspace('d', 0.2, 2.4, 200, unit='Å')
    y_bkg = background(x, bkg_a0=a0, bkg_a1=a1) + sc.array(
        dims=x.dims, values=rng.normal(0.0, 0.3, x.shape), unit='counts'
    )
    y = (
        y_bkg
        + peak(x, peak_amplitude=amplitude1, peak_loc=loc1, peak_scale=scale1)
        + peak(x, peak_amplitude=amplitude2, peak_loc=loc2, peak_scale=scale2)
    )
    with_peak = sc.DataArray(y, coords={'d': x})
    original = with_peak.copy()
    expected = sc.DataArray(y_bkg, coords={'d': x})

    result1 = FitResult(
        aic=sc.scalar(np.nan),
        assessment=FitAssessment.success,
        background=background,
        message='',
        p_value=sc.scalar(np.nan),
        peak=peak,
        popt=params1,
        red_chisq=sc.scalar(np.nan),
        window=sc.array(dims=['range'], values=[1.6, 2.2], unit='Å'),
    )
    result2 = FitResult(
        aic=sc.scalar(np.nan),
        assessment=FitAssessment.success,
        background=background,
        message='',
        p_value=sc.scalar(np.nan),
        peak=peak,
        popt=params2,
        red_chisq=sc.scalar(np.nan),
        window=sc.array(dims=['range'], values=[0.7, 1.1], unit='Å'),
    )
    without_peak = remove_peaks(with_peak, [result1, result2])

    sc.testing.assert_identical(with_peak, original)
    sc.testing.assert_allclose(without_peak, expected, rtol=sc.scalar(1e-5))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

# flake8: noqa
r"""
Specifics for time-of-flight neutron-scattering data reduction, including coordinate transformations.
"""

from . import chopper_cascade, unwrap
from .diagram import TimeDistanceDiagram

__all__ = [
    'chopper_cascade',
    'unwrap',
    'TimeDistanceDiagram',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import uuid

import numpy as np
import scipp as sc


def find_plateaus(
    data: sc.DataArray,
    *,
    atol: sc.Variable,
    min_n_points: int | sc.Variable,
    plateau_dim: str = "plateau",
) -> sc.DataArray:
    """Find regions where the input data is approximately constant.

    Plateaus are found by collecting streaks of points where the derivative of the input
    is less than the tolerance.
    Variances are ignored for this comparison.

    Warning
    -------
    This function technically does not search for plateaus,
    but regions with small derivative.
    This means that if there is a slope smaller than the noise in the input data,
    that sloped region may be falsely identified as a plateau.
    ``find_plateaus`` attempts to catch such a case and raise a ``RuntimeError``,
    but you should always inspect the result!

    Parameters
    ----------
    data:
        Data to search for plateaus in.
        Must be 1-dimensional.
    atol:
        Absolute tolerance for the derivative.
    min_n_points:
        Only return plateaus that have at least this many points.
    plateau_dim:
        Name for the plateau-dimension in the output.

    Returns
    -------
    :
        A 1d binned data array where each bin holds the data points for one plateau.

    See Also
    --------
    scippneutron.chopper.filtering.collapse_plateaus:
        Combine plateau bins into dense data.
    """
    if data.ndim != 1:
        raise NotImplementedError(
            "find_plateaus only supports 1-dimensional data, " f"got {data.ndim} dims"
        )
    if not sc.issorted(data.coords[data.dim], data.dim, order="ascending"):
        raise sc.CoordError(
            "The coord used by find_plateaus must be sorted in ascending order"
        )

    min_n_points = (
        min_n_points
        if isinstance(min_n_points, sc.Variable)
        else sc.index(min_n_points)
    )
    derivative = _derive(data)
    group_id = sc.cumsum(
        (abs(derivative) > atol.to(unit=derivative.unit)).to(dtype="int64")
    )
    # Prepend a 0 to align the groups with the data points (diff reduces length by 1).
    group_id = sc.concat([sc.index(0, dtype="int64"), group_id], dim=derivative.dim)

    group_label = str(uuid.uuid4())
    to_group = data.copy(deep=False)
    to_group.coords[group_label] = group_id
    groups = to_group.group(group_label)
    del groups.coords[group_label]

    plateaus = groups[groups.bins.size().data >= min_n_points].rename_dims(
        {group_label: plateau_dim}
    )
    plateaus.coords[plateau_dim] = sc.arange(plateau_dim, len(plateaus), unit=None)
    if exceeds_tolerance := _check_total_tolerance(plateaus, atol=atol):
        raise RuntimeError(
            f"The following plateaus exceed the tolerance: {exceeds_tolerance}"
        )
    return plateaus


def _derive(da: sc.DataArray) -> sc.Variable:
    x = da.coords[da.dim]
    y = da.data
    return (y[1:] - y[:-1]) / (x[1:] - x[:-1])


def _check_total_tolerance(plateaus: sc.DataArray, *, atol: sc.Variable) -> list[int]:
    # We assume that the noise within a plateau is random.
    # So if the points within a plateau were reordered arbitrarily, the slopes between
    # all neighbors must still be within tolerance.
    # So this function takes the extreme points of each plateau, pretends they
    # are next to each other and computes the corresponding slope and compares
    # it to the tolerance.
    #
    # There is a fudge factor of 2 in the comparison to allow for some larger
    # deviations, especially around the ends of a plateau.
    # Without it, the check would almost always fail.
    exceeds_tolerance = []
    for plateau_bin in plateaus:
        plateau = plateau_bin.value
        max_diff = plateau.data.max() - plateau.data.min()
        coord = plateau.coords[plateau.dim]
        average_step = sc.mean(coord[1:] - coord[:-1])
        slope = max_diff / average_step
        if slope > 2 * atol.to(unit=slope.unit):
            exceeds_tolerance.append(plateau_bin.coords[plateaus.dim].value)
    return exceeds_tolerance


def _next_highest(x: sc.Variable) -> sc.Variable:
    if x.dtype in ("float64", "float32"):
        return sc.array(
            dims=x.dims,
            variances=x.variances,
            unit=x.unit,
            dtype=x.dtype,
            values=np.nextafter(x.values, np.inf),
        )
    if x.dtype == "datetime64":
        return x + sc.scalar(1, dtype="int64", unit=x.unit)
    return x + sc.scalar(1, dtype=x.dtype, unit=x.unit)


def collapse_plateaus(plateaus: sc.DataArray, *, coord: str = "time") -> sc.DataArray:
    """Merge plateaus bins into dense data.

    Useful for post-processing the result of
    :func:`scippneutron.chopper.find_plateaus
    <scippneutron.chopper.filtering.find_plateaus>`.
    Averages the bins and turns the given event coordinate into a bin-edge coordinate.

    Parameters
    ----------
    plateaus:
        Data that has been binned into plateaus.
    coord:
        Name of the constructed bin-edge coordinate.

    Returns
    -------
    :
        Dense data with one element per plateau.

    See Also
    --------
    scippneutron.chopper.filtering.find_plateaus:
        Function to construct suitable input data.
    """
    collapsed = plateaus.bins.mean()
    low = plateaus.bins.coords[coord].bins.min()
    high = _next_highest(plateaus.bins.coords[coord].bins.max())
    collapsed.coords[coord] = sc.concat([low, high], dim=coord).transpose()
    return collapsed


def _is_approximate_multiple(
    x: sc.Variable, *, ref: sc.Variable, rtol: sc.Variable
) -> sc.Variable:
    # If x = n * ref
    quot = x / ref
    a = abs(sc.round(quot) - quot) < rtol
    # If n = x * ref
    quot = sc.reciprocal(quot)
    b = abs(sc.round(quot) - quot) < rtol
    return a | b


def _is_in_phase(
    frequency: sc.DataArray, *, reference: sc.Variable, rtol: sc.Variable
) -> sc.Variable:
    return _is_approximate_multiple(frequency.data, ref=reference, rtol=rtol)


def filter_in_phase(
    frequency: sc.DataArray, *, reference: sc.Variable, rtol: sc.Variable
) -> sc.DataArray:
    """Remove all elements where a given frequency is not in phase with a reference.

    Frequencies are considered to be in-phase if they are an integer multiple
    of the reference or vice versa.

    Parameters
    ----------
    frequency:
        Frequency data to filter.
    reference:
        Reference frequency, e.g., the neutron source frequency.
    rtol:
        Relative tolerance.

    Returns
    -------
    :
        ``frequency`` with all out-of-phase elements removed.
    """
    in_phase = _is_in_phase(frequency, reference=reference, rtol=rtol)
    return frequency[in_phase]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import importlib.resources
from functools import lru_cache
from string import Template

try:
    from scipp.visualization.resources import _preprocess_style, load_dg_style
except ImportError:
    from scipp.html.resources import _preprocess_style, load_dg_style


def _read_text(filename: str, group: str) -> str:
    if hasattr(importlib.resources, 'files'):
        # Use new API added in Python 3.9
        return (
            importlib.resources.files(f'scippneutron._html_repr.{group}')
            .joinpath(filename)
            .read_text()
        )
    # Old API, deprecated as of Python 3.11
    return importlib.resources.read_text(f'scippneutron._html_repr.{group}', filename)


@lru_cache(maxsize=1)
def disk_chopper_repr_template() -> Template:
    return Template(_read_text('disk_chopper_repr.html.template', 'templates'))


@lru_cache(maxsize=1)
def disk_chopper_style() -> str:
    style = _preprocess_style(_read_text('disk_chopper.css', 'styles'))
    return load_dg_style() + f'<style>{style}</style>'
